"""
Sink module for `subprocess`
"""

import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import on_import, patch_function, before


def try_join_iterable(iterable):
    try:
        return " ".join(iterable)
    except Exception:
        return None


@before
def _subprocess_init(func, instance, args, kwargs):
    shell_arguments = get_argument(args, kwargs, 0, "args")
    shell_enabled = get_argument(args, kwargs, 8, "shell")
    if not shell_enabled:
        return  # default shell property is False, we only want to scan if it's True

    command = try_join_iterable(shell_arguments)
    if isinstance(shell_arguments, str):
        command = shell_arguments
    if not command:
        return

    op = "subprocess.Popen"
    register_call(op, "exec_op")

    vulns.run_vulnerability_scan(
        kind="shell_injection",
        op=op,
        args=(command,),
    )


@on_import("subprocess")
def patch(m):
    """
    patching subprocess module
    - patches Popen.__init__ constructor
    - does not patch: check_output, check_call, call, and run (call Popen class)
    """
    patch_function(m, "Popen.__init__", _subprocess_init)



"""Exports function extract_ip_array_from_results"""


def extract_ip_array_from_results(dns_results):
    """Extracts IP's from dns results"""
    ip_addresses = []
    for result in dns_results:
        ip_object = result[4]
        if ip_object is not None and ip_object[0] is not None:
            ip_addresses.append(ip_object[0])
    return ip_addresses



import asyncio
import pytest
from unittest.mock import patch
import aikido_zen.sinks.pymongo
from aikido_zen.background_process.comms import reset_comms
import pymongo.errors as mongo_errs


@pytest.fixture
def db():
    from motor.motor_asyncio import AsyncIOMotorClient

    client = AsyncIOMotorClient("mongodb://admin:password@127.0.0.1:27017")
    return client["my_database"]


@pytest.mark.asyncio
async def test_replace_one(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        filter, repl = {"dog_name": "test"}, {"dog_name": "dog2", "pswd": "pswd"}
        await dogs.replace_one(filter, repl)

        called_with = mock_run_vulnerability_scan.call_args[1]
        mock_run_vulnerability_scan.assert_any_call(
            args=({"dog_name": "test"},),
            op="pymongo.collection.Collection.replace_one",
            kind="nosql_injection",
        )


@pytest.mark.asyncio
async def test_update_one(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        filter, update = {"dog_name": "test"}, {"pswd": "pswd"}
        await dogs.update_one(filter, {"$set": update})

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == filter
        assert called_with["op"] == "pymongo.collection.Collection.update_one"
        assert called_with["kind"] == "nosql_injection"


@pytest.mark.asyncio
async def test_update_many(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        filter, update = {"dog_name": "test"}, {"pswd": "pswd"}
        await dogs.update_many(filter, {"$set": update})

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == filter
        assert called_with["op"] == "pymongo.collection.Collection.update_many"
        assert called_with["kind"] == "nosql_injection"


@pytest.mark.asyncio
async def test_delete_one(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        filter = {"dog_name": "test"}
        await dogs.delete_one(filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == filter
        assert called_with["op"] == "pymongo.collection.Collection.delete_one"
        assert called_with["kind"] == "nosql_injection"


@pytest.mark.asyncio
async def test_delete_many(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        filter = {"dog_name": "test"}
        await dogs.delete_many(filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == filter
        assert called_with["op"] == "pymongo.collection.Collection.delete_many"
        assert called_with["kind"] == "nosql_injection"


@pytest.mark.asyncio
async def test_find_one(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        await dogs.find_one(_filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.find"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_count_documents(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        await dogs.count_documents(_filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.count_documents"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_find_one_and_delete(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        await dogs.find_one_and_delete(_filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.find_one_and_delete"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_find_one_and_replace(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        await dogs.find_one_and_replace(
            filter=_filter, replacement={"dog_name": "test2"}
        )

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.find_one_and_replace"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_find_one_and_update(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        await dogs.find_one_and_update(_filter, {"$set": {"dog_name": "test2"}})

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.find_one_and_update"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_find_empty(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        dogs.find()
        mock_run_vulnerability_scan.assert_not_called()


@pytest.mark.asyncio
async def test_find_not_empty(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        dogs.find(_filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.find"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_find_raw_batches(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        dogs.find_raw_batches(_filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.find_raw_batches"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_distinct(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        await dogs.distinct("pswd", _filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.distinct"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_distinct_kwargs(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        _filter = {"dog_name": "test", "pswd": "pswd"}
        await dogs.distinct(key="pswd", filter=_filter)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == _filter
        assert called_with["op"] == "pymongo.collection.Collection.distinct"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_distinct_empty(db):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        await dogs.distinct("pswd")

        mock_run_vulnerability_scan.assert_not_called()


@pytest.mark.asyncio
async def test_aggregate(db):
    pipeline = [
        {
            "$group": {
                "_id": "$item",
                "total_quantity": {"$sum": "$quantity"},
                "average_price": {"$avg": "$price"},
            }
        },
        {"$sort": {"total_quantity": -1}},  # Sort by total_quantity in descending order
    ]
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        async for doc in dogs.aggregate(pipeline):
            pass

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == pipeline
        assert called_with["op"] == "pymongo.collection.Collection.aggregate"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_aggregate_key(db):
    pipeline = [
        {
            "$group": {
                "_id": "$item",
                "total_quantity": {"$sum": "$quantity"},
                "average_price": {"$avg": "$price"},
            }
        },
        {"$sort": {"total_quantity": -1}},  # Sort by total_quantity in descending order
    ]
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        async for doc in dogs.aggregate(pipeline=pipeline):
            pass

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == pipeline
        assert called_with["op"] == "pymongo.collection.Collection.aggregate"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_aggregate_raw_batches_key(db):
    pipeline = [
        {
            "$group": {
                "_id": "$item",
                "total_quantity": {"$sum": "$quantity"},
                "average_price": {"$avg": "$price"},
            }
        },
        {"$sort": {"total_quantity": -1}},  # Sort by total_quantity in descending order
    ]
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        async for doc in dogs.aggregate_raw_batches(pipeline=pipeline):
            pass

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == pipeline
        assert (
            called_with["op"] == "pymongo.collection.Collection.aggregate_raw_batches"
        )
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_aggregate_raw_batches(db):
    pipeline = [
        {
            "$group": {
                "_id": "$item",
                "total_quantity": {"$sum": "$quantity"},
                "average_price": {"$avg": "$price"},
            }
        },
        {"$sort": {"total_quantity": -1}},  # Sort by total_quantity in descending order
    ]
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        async for doc in dogs.aggregate_raw_batches(pipeline):
            pass

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == pipeline
        assert (
            called_with["op"] == "pymongo.collection.Collection.aggregate_raw_batches"
        )
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_watch(db):
    pipeline = [
        {"$match": {"operationType": "insert"}},  # Only listen for insert operations
        {
            "$project": {
                "fullDocument": 1,  # Include the full document in the output
                "operationType": 1,  # Include the operation type
            }
        },
    ]
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        with pytest.raises(mongo_errs.OperationFailure):
            async with dogs.watch(pipeline) as change_stream:
                async for change in change_stream:
                    pass

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == pipeline
        assert called_with["op"] == "pymongo.collection.Collection.watch"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_watch_key(db):
    pipeline = [
        {
            "$group": {
                "_id": "$item",
                "total_quantity": {"$sum": "$quantity"},
                "average_price": {"$avg": "$price"},
            }
        },
        {"$sort": {"total_quantity": -1}},  # Sort by total_quantity in descending order
    ]
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        with pytest.raises(mongo_errs.OperationFailure):
            async with dogs.watch(pipeline=pipeline) as change_stream:
                pass

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == pipeline
        assert called_with["op"] == "pymongo.collection.Collection.watch"
        assert called_with["kind"] == "nosql_injection"
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.asyncio
async def test_bulk_write(db):
    from pymongo import InsertOne, UpdateOne, DeleteOne

    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        dogs = db["dogs"]
        requests = [
            InsertOne({"dog_name": "Buddy2", "age": 3}),
            InsertOne({"dog_name": "Max3"}),
            UpdateOne({"dog_name": "Buddy1"}, {"$set": {"age": 0}}),
            DeleteOne({"dog_name": "Max3"}),
            InsertOne({"dog_name": "Bella4"}),
        ]
        try:
            await dogs.bulk_write(requests)
        except Exception:
            pass
        calls = mock_run_vulnerability_scan.call_args_list

        called_with1 = calls[0][1]
        assert called_with1["args"][0] == {"dog_name": "Buddy1"}
        assert called_with1["op"] == "pymongo.collection.Collection.bulk_write"
        assert called_with1["kind"] == "nosql_injection"

        called_with2 = calls[1][1]
        assert called_with2["args"][0] == {"dog_name": "Max3"}
        assert called_with2["op"] == "pymongo.collection.Collection.bulk_write"
        assert called_with2["kind"] == "nosql_injection"



"""
Sink module for `pymysql`
"""

import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import patch_function, on_import, before


@before
def _execute(func, instance, args, kwargs):
    query = get_argument(args, kwargs, 0, "query")
    if isinstance(query, bytearray):
        # If query is type bytearray, it will be picked up by our wrapping of executemany
        return

    register_call("pymysql.Cursor.execute", "sql_op")
    vulns.run_vulnerability_scan(
        kind="sql_injection", op="pymysql.Cursor.execute", args=(query, "mysql")
    )


@before
def _executemany(func, instance, args, kwargs):
    query = get_argument(args, kwargs, 0, "query")

    register_call("pymysql.Cursor.executemany", "sql_op")
    vulns.run_vulnerability_scan(
        kind="sql_injection", op="pymysql.Cursor.executemany", args=(query, "mysql")
    )


@on_import("pymysql.cursors", "pymysql", version_requirement="0.9.0")
def patch(m):
    """
    patching `pymysql.cursors`
    - patches Cursor.execute(query)
    - patches Cursor.executemany(query)
    https://github.com/PyMySQL/PyMySQL/blob/95635f587ba9076e71a223b113efb08ac34a361d/pymysql/cursors.py#L133
    """
    patch_function(m, "Cursor.execute", _execute)
    patch_function(m, "Cursor.executemany", _executemany)



"""
Aikido background process, this will create a new process
and listen for data sent by our sources and sinks
"""

import os
from multiprocessing import Process
import platform

from aikido_zen.helpers.token import get_token_from_env
from aikido_zen.helpers.get_temp_dir import get_temp_dir
from aikido_zen.helpers.hash_aikido_token import hash_aikido_token
from aikido_zen.helpers.logging import logger

from aikido_zen.background_process.comms import (
    AikidoIPCCommunications,
    get_comms,
    reset_comms,
)
from .aikido_background_process import AikidoBackgroundProcess


def start_background_process():
    """
    Starts a process to handle incoming/outgoing data
    """

    # Generate a secret key :
    secret_key_bytes = str.encode(str(get_token_from_env()))
    if secret_key_bytes == b"None":
        logger.warning(
            "You are running without AIKIDO_TOKEN set, not starting background process."
        )
        return

    address = get_uds_filename()
    if platform.system() == "Windows":
        # Python does not support Windows UDS just yet, so we have to rely on INET
        address = ("127.0.0.1", 49156)

    comms = AikidoIPCCommunications(address, secret_key_bytes)

    if isinstance(address, str) and os.path.exists(address):
        if background_process_already_active(comms):
            return  # Don't start if the background process is already active
        os.remove(address)

    #  Daemon is set to True so that the process kills itself when the main process dies
    background_process = Process(
        target=AikidoBackgroundProcess, args=(comms.address, comms.key), daemon=True
    )
    background_process.start()


def get_uds_filename():
    """Returns the address for UDS file"""
    temp_dir = get_temp_dir()
    prefix = "aikido_python"
    name = hash_aikido_token()
    return f"{temp_dir}/{prefix}_{name}.sock"


def background_process_already_active(comms):
    res = comms.send_data_to_bg_process(action="PING", obj=tuple(), receive=True)
    if res["success"] and res["data"] == "Received":
        # Ping is active, return.
        logger.debug(
            "A background agent is already running, not starting a new background agent."
        )
        return True
    return False



"""Mainly exports on_start function"""

from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms


def on_start(connection_manager):
    """
    This will send out an Event signalling the start to the server
    """
    if not connection_manager.token:
        return
    res = connection_manager.api.report(
        connection_manager.token,
        {
            "type": "started",
            "time": get_unixtime_ms(),
            "agent": connection_manager.get_manager_info(),
        },
        connection_manager.timeout_in_sec,
    )
    if not res.get("success", True):
        # Update config time even in failure :
        connection_manager.conf.last_updated_at = get_unixtime_ms()
        logger.error("Failed to communicate with Aikido Server : %s", res["error"])
    else:
        connection_manager.update_service_config(res)
        connection_manager.update_firewall_lists()
        logger.info("Established connection with Aikido Server")
    return res



"""
Provides all the functionality for contexts
"""

import contextvars
import json
from json import JSONDecodeError
from time import sleep
from urllib.parse import parse_qs

from aikido_zen.helpers.build_route_from_url import build_route_from_url
from aikido_zen.helpers.get_subdomains_from_url import get_subdomains_from_url
from aikido_zen.helpers.logging import logger
from .wsgi import set_wsgi_attributes_on_context
from .asgi import set_asgi_attributes_on_context
from .extract_route_params import extract_route_params

UINPUT_SOURCES = ["body", "cookies", "query", "headers", "xml", "route_params"]
current_context = contextvars.ContextVar("current_context", default=None)

WSGI_SOURCES = ["django", "flask"]
ASGI_SOURCES = ["quart", "django_async", "starlette"]


def get_current_context():
    """Returns the current context"""
    try:
        return current_context.get()
    except Exception:
        return None


class Context:
    """
    A context object, it stores everything that is important
    for vulnerability detection
    """

    def __init__(self, context_obj=None, body=None, req=None, source=None):
        if context_obj:
            logger.debug("Creating Context instance based on dict object.")
            self.__dict__.update(context_obj)
            return
        # Define emtpy variables/Properties :
        self.source = source
        self.user = None
        self.parsed_userinput = {}
        self.xml = {}
        self.outgoing_req_redirects = []
        self.set_body(body)

        # Parse WSGI/ASGI/... request :
        self.cookies = self.method = self.remote_address = self.query = self.headers = (
            self.url
        ) = None
        if source in WSGI_SOURCES:
            set_wsgi_attributes_on_context(self, req)
        elif source in ASGI_SOURCES:
            set_asgi_attributes_on_context(self, req)

        # Define variables using parsed request :
        self.route = build_route_from_url(self.url)
        self.route_params = extract_route_params(self.url)
        self.subdomains = get_subdomains_from_url(self.url)

        self.executed_middleware = False

    def __reduce__(self):
        return (
            self.__class__,
            (
                {
                    "method": self.method,
                    "remote_address": self.remote_address,
                    "url": self.url,
                    "body": self.body,
                    "headers": self.headers,
                    "query": self.query,
                    "cookies": self.cookies,
                    "source": self.source,
                    "route": self.route,
                    "subdomains": self.subdomains,
                    "user": self.user,
                    "xml": self.xml,
                    "outgoing_req_redirects": self.outgoing_req_redirects,
                    "executed_middleware": self.executed_middleware,
                    "route_params": self.route_params,
                },
                None,
                None,
            ),
        )

    def set_as_current_context(self):
        """
        Set the current context
        """
        current_context.set(self)

    def set_body(self, body):
        try:
            self.set_body_internal(body)
        except Exception as e:
            logger.debug("Exception occurred whilst setting body: %s", e)

    def set_body_internal(self, body):
        """Sets the body and checks if it's possibly JSON"""
        self.body = body
        if isinstance(self.body, (str, bytes)) and len(body) == 0:
            # Make sure that empty bodies like b"" don't get sent.
            self.body = None
        if isinstance(self.body, bytes):
            self.body = self.body.decode("utf-8")  # Decode byte input to string.
        if not isinstance(self.body, str):
            return
        if self.body.strip()[0] in ["{", "[", '"']:
            # Might be JSON, but might not have been parsed correctly by server because of wrong headers
            parsed_body = json.loads(self.body)
            if parsed_body:
                self.body = parsed_body

    def get_route_metadata(self):
        """Returns a route_metadata object"""
        return {
            "method": self.method,
            "route": self.route,
            "url": self.url,
        }

    def get_user_agent(self):
        if "USER_AGENT" in self.headers:
            return self.headers["USER_AGENT"]
        return None



import pytest
from .is_localhost_ip import is_localhost_ip


def test_is_localhost_ip():
    assert is_localhost_ip("127.0.0.1") == True
    assert is_localhost_ip("::ffff:127.0.0.1") == True
    assert is_localhost_ip("::1") == True
    assert is_localhost_ip("192.168.1.1") == False
    assert is_localhost_ip("localhost") == False  # Not a valid IP format



import pytest
from unittest.mock import patch
import aikido_zen.sinks.asyncpg
import aikido_zen.sinks.os
from aikido_zen.background_process.comms import reset_comms

kind = "sql_injection"
op = "MySQLdb.connections.query"


@pytest.fixture
async def database_conn():
    import asyncpg

    conn = await asyncpg.connect(
        host="127.0.0.1", user="user", password="password", database="db"
    )
    return conn


@pytest.mark.asyncio
async def test_conn_execute(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        conn = await database_conn
        query = "SELECT * FROM dogs"
        await conn.execute(query)

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        called_with_op = mock_run_vulnerability_scan.call_args[1]["op"]
        called_with_kind = mock_run_vulnerability_scan.call_args[1]["kind"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"
        assert called_with_op == "asyncpg.connection.Connection.execute"
        assert called_with_kind == "sql_injection"

        await conn.close()


@pytest.mark.asyncio
async def test_conn_fetchrow(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        conn = await database_conn
        query = "SELECT * FROM dogs"
        await conn.fetchrow(query)

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"

        await conn.close()


@pytest.mark.asyncio
async def test_conn_fetch(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        conn = await database_conn
        query = "SELECT * FROM dogs"
        await conn.fetch(query)

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"

        await conn.close()


@pytest.mark.asyncio
async def test_conn_fetchval(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        conn = await database_conn
        query = "SELECT COUNT(*) FROM dogs"
        await conn.fetchval(query)

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"

        await conn.close()


@pytest.mark.asyncio
async def test_conn_execute_parameterized(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        conn = await database_conn
        query = "INSERT INTO dogs(dog_name, isadmin) VALUES($1, $2)"
        await conn.execute(query, "Doggo", False)
        calls = mock_run_vulnerability_scan.call_args_list

        counter = 0
        for call in calls:
            args = call[1]["args"]
            if args[0] == query:
                counter += 1
        assert counter == 2

        await conn.close()


@pytest.mark.asyncio
async def test_conn_transaction(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        conn = await database_conn
        async with conn.transaction():
            query = "SELECT COUNT(*) FROM dogs"
            await conn.execute(query)

            calls = mock_run_vulnerability_scan.call_args_list
            begin_in_calls = False
            query_in_calls = False
            for call in calls:
                args = call[1]["args"]
                if args[0] == "BEGIN;":
                    begin_in_calls = True
                if args[0] == query:
                    query_in_calls = True
            assert begin_in_calls == query_in_calls == True

        await conn.close()


@pytest.mark.asyncio
async def test_conn_cursor(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        conn = await database_conn
        async with conn.transaction():
            query = "SELECT COUNT(*) FROM dogs"
            await conn.cursor(query)

            calls = mock_run_vulnerability_scan.call_args_list

            called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
            assert called_with_args[0] == "BEGIN;"
            assert called_with_args[1] == "postgres"

        await conn.close()



import pytest
import signal
from unittest.mock import patch
import aikido_zen.sources.flask
from aikido_zen.background_process.comms import reset_comms
from aikido_zen.context import current_context, get_current_context

sample_environ = {
    "REQUEST_METHOD": "POST",
    "HTTP_HEADER_1": "header 1 value",
    "HTTP_HEADER_2": "Header 2 value",
    "RANDOM_VALUE": "Random value",
    "HTTP_COOKIE": "sessionId=abc123xyz456;",
    "wsgi.url_scheme": "https",
    "HTTP_HOST": "example.com",
    "PATH_INFO": "/hello",
    "QUERY_STRING": "user=JohnDoe&age=30&age=35",
    "body": "username=JohnDoe&age=30&age=35",
    "CONTENT_TYPE": "application/x-www-form-urlencoded",
    "REMOTE_ADDR": "198.51.100.23",
}

sample_environ_malformed_cookie = {
    "REQUEST_METHOD": "POST",
    "HTTP_HEADER_1": "header 1 value",
    "HTTP_HEADER_2": "Header 2 value",
    "RANDOM_VALUE": "Random value",
    "HTTP_COOKIE": "\u0000" * 10,
    "wsgi.url_scheme": "https",
    "HTTP_HOST": "example.com",
    "PATH_INFO": "/hello",
    "QUERY_STRING": "user=JohnDoe&age=30&age=35",
    "body": '{"asd": 1}',
    "CONTENT_TYPE": "application/x-www-form-urlencoded",
    "REMOTE_ADDR": "198.51.100.23",
}

sample_environ_malformed_json = {
    "REQUEST_METHOD": "POST",
    "HTTP_HEADER_1": "header 1 value",
    "HTTP_HEADER_2": "Header 2 value",
    "RANDOM_VALUE": "Random value",
    "HTTP_COOKIE": "sessionId=abc123xyz456;",
    "wsgi.url_scheme": "https",
    "HTTP_HOST": "example.com",
    "PATH_INFO": "/hello",
    "body": '{"asd": 1invalidjson}',
    "QUERY_STRING": "user=JohnDoe",
    "CONTENT_TYPE": "application/json",
    "REMOTE_ADDR": "198.51.100.23",
}

sample_environ_view_args_and_malformed_json = {
    "REQUEST_METHOD": "POST",
    "HTTP_COOKIE": "sessionId=abc123xyz456;",
    "HTTP_HEADER_1": "header 1 value",
    "HTTP_HEADER_2": "Header 2 value",
    "HTTP_HOST": "example.com",
    "CONTENT_TYPE": "application/json",
    "PATH_INFO": "/hello/JohnDoe/30",
    "QUERY_STRING": "",
    "body": '{"invalid_json": true',
    "REMOTE_ADDR": "198.51.100.23",
    "wsgi.url_scheme": "https",
}


class TimeoutException(Exception):
    pass


def timeout_handler(signum, frame):
    raise TimeoutException


signal.signal(signal.SIGALRM, timeout_handler)


def test_flask_all_3_func_with_view_args_and_invalid_json_body():
    with patch(
        "aikido_zen.sources.functions.request_handler.request_handler"
    ) as mock_request_handler:
        reset_comms()
        current_context.set(None)
        mock_request_handler.return_value = None

        from flask import Flask

        app = Flask(__name__)

        @app.route("/hello/<user>/<age>", methods=["POST"])
        def hello(user, age):
            return f"User: {user}, Age: {age}"

        try:
            signal.alarm(1)

            app(sample_environ_view_args_and_malformed_json, lambda x, y: x)
            app.run()

        except TimeoutException:
            pass

        assert get_current_context().method == "POST"
        assert get_current_context().body is None
        assert get_current_context().headers == {
            "COOKIE": "sessionId=abc123xyz456;",
            "HEADER_1": "header 1 value",
            "HEADER_2": "Header 2 value",
            "HOST": "example.com",
            "CONTENT_TYPE": "application/json",
        }
        calls = mock_request_handler.call_args_list
        assert len(calls) == 3
        assert calls[0][1]["stage"] == "init"
        assert calls[1][1]["stage"] == "pre_response"
        assert calls[2][1]["stage"] == "post_response"
        assert calls[2][1]["status_code"] == 200

        assert get_current_context().route_params["user"] == "JohnDoe"
        assert get_current_context().route_params["age"] == "30"


def test_flask_all_3_func_with_malformed_cookie():
    """When the flask body can not be parsed (because it contains invalid json for example), we should still parse the cookies of the endpoint"""
    with patch(
        "aikido_zen.sources.functions.request_handler.request_handler"
    ) as mock_request_handler:
        reset_comms()
        current_context.set(None)
        mock_request_handler.return_value = None

        from flask import Flask

        app = Flask(__name__)
        try:
            signal.alarm(1)
            app(sample_environ_malformed_cookie, lambda x, y: x)
            app.run()
        except TimeoutException:
            pass

        print(get_current_context())

        assert get_current_context().method == "POST"
        assert get_current_context().cookies == {"\u0000" * 10: ""}

        calls = mock_request_handler.call_args_list
        assert len(calls) == 3
        assert calls[0][1]["stage"] == "init"
        assert calls[1][1]["stage"] == "pre_response"
        assert calls[2][1]["stage"] == "post_response"
        assert calls[2][1]["status_code"] == 404


def test_flask_all_3_func_with_invalid_body():
    """When the flask body can not be parsed (because it contains invalid json for example), we should still parse the cookies of the endpoint"""
    with patch(
        "aikido_zen.sources.functions.request_handler.request_handler"
    ) as mock_request_handler:
        reset_comms()
        current_context.set(None)
        mock_request_handler.return_value = None

        from flask import Flask

        app = Flask(__name__)
        try:
            signal.alarm(1)
            app(sample_environ_malformed_json, lambda x, y: x)
            app.run()
        except TimeoutException:
            pass

        assert get_current_context().method == "POST"
        assert (
            get_current_context().body == None
        )  # body is None since it's invalid json
        assert get_current_context().headers == {
            "COOKIE": "sessionId=abc123xyz456;",
            "HEADER_1": "header 1 value",
            "HEADER_2": "Header 2 value",
            "HOST": "example.com",
            "CONTENT_TYPE": "application/json",
        }
        calls = mock_request_handler.call_args_list
        assert len(calls) == 3
        assert calls[0][1]["stage"] == "init"
        assert calls[1][1]["stage"] == "pre_response"
        assert calls[2][1]["stage"] == "post_response"
        assert calls[2][1]["status_code"] == 404


def test_flask_all_3_func():
    with patch(
        "aikido_zen.sources.functions.request_handler.request_handler"
    ) as mock_request_handler:
        reset_comms()
        current_context.set(None)
        mock_request_handler.return_value = None

        from flask import Flask

        app = Flask(__name__)
        try:
            signal.alarm(1)
            app(sample_environ, lambda x, y: x)
            app.run()
        except TimeoutException:
            pass
        assert get_current_context().method == "POST"
        assert get_current_context().body == None
        assert get_current_context().headers == {
            "COOKIE": "sessionId=abc123xyz456;",
            "HEADER_1": "header 1 value",
            "HEADER_2": "Header 2 value",
            "HOST": "example.com",
            "CONTENT_TYPE": "application/x-www-form-urlencoded",
        }
        calls = mock_request_handler.call_args_list
        assert len(calls) == 3
        assert calls[0][1]["stage"] == "init"
        assert calls[1][1]["stage"] == "pre_response"
        assert calls[2][1]["stage"] == "post_response"
        assert calls[2][1]["status_code"] == 404


def test_startup_flask():
    reset_comms()
    from flask import Flask

    app = Flask(__name__)
    try:
        signal.alarm(1)
        app.run()
    except TimeoutException:
        pass



import pytest
from urllib.parse import urlparse, urlunparse
from .get_redirect_origin import get_redirect_origin


# Helper function to create URL objects
def create_url(href):
    return urlparse(href)


# Test cases
def test_get_redirect_origin():
    assert get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://hackers.com"),
            },
        ],
        "hackers.com",
        443,
    ) == create_url("https://example.com")

    assert get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com/2"),
            },
            {
                "source": create_url("https://example.com/2"),
                "destination": create_url("https://hackers.com/test"),
            },
        ],
        "hackers.com",
        443,
    ) == create_url("https://example.com")


def test_get_redirect_origin_no_redirects():
    assert (
        get_redirect_origin(
            [],
            "hackers.com",
            443,
        )
        is None
    )
    assert (
        get_redirect_origin(
            None,
            "hackers.com",
            443,
        )
        is None
    )


def test_get_redirect_origin_not_a_destination():
    assert (
        get_redirect_origin(
            [
                {
                    "source": create_url("https://example.com"),
                    "destination": create_url("https://hackers.com"),
                },
            ],
            "example.com",
            443,
        )
        is None
    )


def test_get_redirect_origin_not_in_redirects():
    assert (
        get_redirect_origin(
            [
                {
                    "source": create_url("https://example.com"),
                    "destination": create_url("https://hackers.com"),
                },
            ],
            "example.com",
            443,
        )
        is None
    )


def test_get_redirect_origin_multiple_redirects():
    assert get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com/2"),
            },
            {
                "source": create_url("https://example.com/2"),
                "destination": create_url("https://hackers.com/test"),
            },
            {
                "source": create_url("https://hackers.com/test"),
                "destination": create_url("https://another.com"),
            },
        ],
        "hackers.com",
        443,
    ) == create_url("https://example.com")


def test_avoids_infinite_loops_with_unrelated_cyclic_redirects():
    result = get_redirect_origin(
        [
            # Unrelated cyclic redirects
            {
                "source": create_url("https://cycle.com/a"),
                "destination": create_url("https://cycle.com/b"),
            },
            {
                "source": create_url("https://cycle.com/b"),
                "destination": create_url("https://cycle.com/c"),
            },
            {
                "source": create_url("https://cycle.com/c"),
                "destination": create_url("https://cycle.com/a"),
            },
            # Relevant redirects
            {
                "source": create_url("https://start.com"),
                "destination": create_url("https://middle.com"),
            },
            {
                "source": create_url("https://middle.com"),
                "destination": create_url("https://end.com"),
            },
        ],
        "end.com",
        443,
    )
    assert result == create_url("https://start.com")


def test_handles_multiple_requests_with_overlapping_redirects():
    result = get_redirect_origin(
        [
            # Overlapping redirects
            {
                "source": create_url("https://site1.com"),
                "destination": create_url("https://site2.com"),
            },
            {
                "source": create_url("https://site2.com"),
                "destination": create_url("https://site3.com"),
            },
            {
                "source": create_url("https://site3.com"),
                "destination": create_url("https://site1.com"),  # Cycle
            },
            # Relevant redirects
            {
                "source": create_url("https://origin.com"),
                "destination": create_url("https://destination.com"),
            },
        ],
        "destination.com",
        443,
    )
    assert result == create_url("https://origin.com")


def test_avoids_infinite_loops_when_cycles_are_part_of_the_redirect_chain():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://start.com"),
                "destination": create_url("https://loop.com/a"),
            },
            {
                "source": create_url("https://loop.com/a"),
                "destination": create_url("https://loop.com/b"),
            },
            {
                "source": create_url("https://loop.com/b"),
                "destination": create_url("https://loop.com/c"),
            },
            {
                "source": create_url("https://loop.com/c"),
                "destination": create_url("https://loop.com/a"),  # Cycle here
            },
        ],
        "loop.com",
        443,
    )
    assert result == create_url("https://start.com")


def test_redirects_with_query_parameters():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com?param=value"),
            },
        ],
        "example.com",
        443,
    )
    assert result == create_url("https://example.com")


def test_redirects_with_fragment_identifiers():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com#section"),
            },
        ],
        "example.com",
        443,
    )
    assert result == create_url("https://example.com")


def test_redirects_with_different_protocols():
    result = get_redirect_origin(
        [
            {
                "source": create_url("http://example.com"),
                "destination": create_url("https://example.com"),
            },
        ],
        "example.com",
        443,
    )
    assert result == create_url("http://example.com")


def test_redirects_with_different_ports():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com:8080"),
            },
        ],
        "example.com",
        8080,
    )
    assert result == create_url("https://example.com")


def test_redirects_with_paths():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com/home"),
            },
            {
                "source": create_url("https://example.com/home"),
                "destination": create_url("https://example.com/home/welcome"),
            },
        ],
        "example.com",
        443,
    )
    assert result == create_url("https://example.com")


def test_multiple_redirects_to_same_destination():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://a.com"),
                "destination": create_url("https://d.com"),
            },
            {
                "source": create_url("https://b.com"),
                "destination": create_url("https://d.com"),
            },
            {
                "source": create_url("https://c.com"),
                "destination": create_url("https://d.com"),
            },
        ],
        "d.com",
        443,
    )
    assert result == create_url("https://a.com")


def test_multiple_redirect_paths_to_same_url():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://x.com"),
                "destination": create_url("https://y.com"),
            },
            {
                "source": create_url("https://y.com"),
                "destination": create_url("https://z.com"),
            },
            {
                "source": create_url("https://a.com"),
                "destination": create_url("https://b.com"),
            },
            {
                "source": create_url("https://b.com"),
                "destination": create_url("https://z.com"),
            },
        ],
        "z.com",
        443,
    )
    assert result == create_url("https://x.com")


def test_returns_undefined_when_source_and_destination_are_same_url():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com"),
            },
        ],
        "example.com",
        443,
    )
    assert result is None


def test_handles_very_long_redirect_chains():
    redirects = []
    for i in range(100):
        redirects.append(
            {
                "source": create_url(f"https://example.com/{i}"),
                "destination": create_url(f"https://example.com/{i + 1}"),
            }
        )

    result = get_redirect_origin(redirects, "example.com", 443)
    assert result == create_url("https://example.com/0")


def test_handles_redirects_with_cycles_longer_than_one_redirect():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://a.com"),
                "destination": create_url("https://b.com"),
            },
            {
                "source": create_url("https://b.com"),
                "destination": create_url("https://c.com"),
            },
            {
                "source": create_url("https://c.com"),
                "destination": create_url("https://a.com"),
            },
        ],
        "a.com",
        443,
    )
    assert result is None


def test_handles_redirects_with_different_query_parameters():
    result = get_redirect_origin(
        [
            {
                "source": create_url("https://example.com"),
                "destination": create_url("https://example.com?param=1"),
            },
            {
                "source": create_url("https://example.com?param=1"),
                "destination": create_url("https://example.com?param=2"),
            },
        ],
        "example.com",
        443,
    )
    assert result == create_url("https://example.com")



import pytest
from .inspect_getaddrinfo_result import (
    inspect_getaddrinfo_result,
    get_metadata_for_ssrf_attack,
)


def test_get_metadata_for_ssrf_attack():
    assert get_metadata_for_ssrf_attack("hostname2", None) == {"hostname": "hostname2"}
    assert get_metadata_for_ssrf_attack("hostname3", 443) == {
        "hostname": "hostname3",
        "port": "443",
    }



"""Exports set_wsgi_attributes_on_context"""

from urllib.parse import parse_qs
from aikido_zen.helpers.get_ip_from_request import get_ip_from_request
from aikido_zen.helpers.logging import logger
from .extract_wsgi_headers import extract_wsgi_headers
from .build_url_from_wsgi import build_url_from_wsgi
from ..parse_cookies import parse_cookies


def set_wsgi_attributes_on_context(context, environ):
    """
    This extracts WSGI attributes, described in :
    https://peps.python.org/pep-3333/#environ-variables
    """
    logger.debug("Setting wsgi attributes")

    context.method = environ["REQUEST_METHOD"]
    context.headers = extract_wsgi_headers(environ)
    if "COOKIE" in context.headers:
        context.cookies = parse_cookies(context.headers["COOKIE"])
    else:
        context.cookies = {}
    context.url = build_url_from_wsgi(environ)
    context.query = parse_qs(environ["QUERY_STRING"])
    context.remote_address = get_ip_from_request(
        environ["REMOTE_ADDR"], context.headers
    )

    # Content type is generally not included as a header, do include this as a header to simplify :
    if "CONTENT_TYPE" in environ:
        context.headers["CONTENT_TYPE"] = environ["CONTENT_TYPE"]



"""Exports get_binary_path"""

import platform
import os


def get_binary_path():
    """Returns an absolute path for Rust binary file"""
    current_dir = os.path.dirname(os.path.abspath(__file__))
    lib_path = os.path.join(current_dir, "../../libs", get_file_name())
    return lib_path


def get_file_name():
    """Gives you the file name for the binary based on platform info"""
    os_name = platform.system().lower()
    machine = platform.machine().lower()
    file_name = "libzen_internals_"

    # On macOS, platform.machine() returns "arm64" for Apple Silicon
    # On Linux, platform.machine() returns "aarch64" for ARM64
    if "arm64" in machine or "aarch64" in machine:
        file_name += "aarch64-"
    # On macOS, platform.machine() returns "x86_64" for Intel
    # On Linux, platform.machine() returns "x86_64" for AMD64
    elif "x86_64" in machine or "amd64" in machine:
        file_name += "x86_64-"  # x86_64 or AMD64

    if os_name == "windows":
        file_name += "pc-windows-gnu.dll"  # Windows
    elif os_name == "darwin":
        file_name += "apple-darwin.dylib"  # macOS
    elif os_name in ["linux", "linux2"]:
        file_name += "unknown-linux-gnu.so"  # Linux

    return file_name



""" Exports `check_context_for_nosql_injection`"""

from aikido_zen.helpers.serialize_to_json import serialize_to_json
from . import detect_nosql_injection


def check_context_for_nosql_injection(context, op, _filter):
    """Checks the context object for a nosql injection"""
    results = detect_nosql_injection(context, _filter)
    if not results or not results["injection"]:
        return {}
    return {
        "operation": op,
        "kind": "nosql_injection",
        "source": results["source"],
        "pathToPayload": results["pathToPayload"],
        "metadata": {"filter": serialize_to_json(_filter)},
        "payload": results["payload"],
    }



import pytest
from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms
import time


def test_get_unixtime_ms(monkeypatch):
    # Mock time.time to return a specific timestamp
    monkeypatch.setattr(
        time, "time", lambda: 1633072800.123
    )  # Example timestamp in seconds

    # Calculate the expected result in milliseconds
    expected_result = int(1633072800.123 * 1000)

    assert get_unixtime_ms() == expected_result


def test_get_unixtime_ms_zero(monkeypatch):
    # Mock time.time to return zero
    monkeypatch.setattr(time, "time", lambda: 0.0)

    # The expected result should be 0 milliseconds
    assert get_unixtime_ms() == 0



"""
Mainly exports the `get_ip_from_request` function
"""

import socket
import os
from aikido_zen.helpers.logging import logger


def get_ip_from_request(remote_address, headers):
    """
    Tries and get the IP address from the request, checking for x-forwarded-for
    """
    if headers:
        lower_headers = {key.lower(): value for key, value in headers.items()}
        if "x_forwarded_for" in lower_headers and trust_proxy():
            x_forwarded_for = get_client_ip_from_x_forwarded_for(
                lower_headers["x_forwarded_for"]
            )

            if x_forwarded_for and is_ip(x_forwarded_for):
                return x_forwarded_for

    if remote_address and is_ip(remote_address):
        return remote_address

    return None


def get_client_ip_from_x_forwarded_for(value):
    """
    Fetches the IP out of the X-Forwarder-For headers
    """
    forwarded_ips = [ip.strip() for ip in value.split(",")]

    for ip in forwarded_ips:
        if ":" in ip:
            parts = ip.split(":")
            if len(parts) == 2:
                return parts[0]

    for ip in forwarded_ips:
        if is_ip(ip):
            return ip

    return None


def trust_proxy():
    """
    Checks the enviornment variables for `AIKIDO_TRUST_PROXY`
    """
    return not "AIKIDO_TRUST_PROXY" in os.environ or os.environ[
        "AIKIDO_TRUST_PROXY"
    ] in ["1", "true"]


def is_ip(value):
    """
    Checks if `value` is a vlid IPv4 or IPv6 ip
    """
    try:
        socket.inet_pton(socket.AF_INET, value)  # Check for IPv4
        return True
    except socket.error:
        try:
            socket.inet_pton(socket.AF_INET6, value)  # Check for IPv6
            return True
        except socket.error:
            return False
    except Exception as e:
        logger.debug("is_ip failed because of exception : %s", e)
        return False



"""Exports get_data_schema function"""

from .merge_data_schemas import merge_data_schemas

# Maximum depth to traverse the data structure to get the schema for improved performance
MAX_TRAVERSAL_DEPTH = 20
# Maximum amount of array members to merge into one
MAX_ARRAY_DEPTH = 10
# Maximum number of properties per level
MAX_PROPS = 100


def get_data_schema(data, depth=0):
    """
    Get the schema of the data (for example, HTTP JSON body) as a schema.
    """
    if data is None:
        # Handle None as a special case
        return {"type": "null"}

    # If the data is not an object (or an array), return the type
    if not isinstance(data, (dict, list, tuple)):
        type_normalized = normalize_python_types(type(data).__name__)
        return {"type": type_normalized}

    if isinstance(data, (list, tuple)):
        # Assume that the array is homogenous (for performance reasons)
        items = None
        for i in range(min(MAX_ARRAY_DEPTH, len(data))):
            current_data_schema = get_data_schema(data[i])
            if items is None:
                items = current_data_schema
            else:
                items = merge_data_schemas(items, current_data_schema)

        return {"type": "array", "items": items}

    schema = {"type": "object", "properties": {}}

    # If the depth is less than the maximum depth, get the schema for each property
    if depth < MAX_TRAVERSAL_DEPTH:
        prop_count = 0
        for key, value in data.items():
            if prop_count >= MAX_PROPS:
                # We cannot allow more properties than MAX_PROPS, breaking for loop.
                break
            schema["properties"][key] = get_data_schema(value, depth + 1)
            prop_count += 1

    return schema


def normalize_python_types(type_name):
    """
    Normalizes python types :
    str -> string
    int,float -> number
    bool -> boolean
    """
    if type_name == "str":
        return "string"
    if type_name in ["int", "float"]:
        return "number"
    if type_name == "bool":
        return "boolean"
    return type_name



import pytest
from unittest.mock import patch
import aikido_zen.sinks.builtins

kind = "path_traversal"


def test_gzip_gzipfile():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import gzip

        with pytest.raises(FileNotFoundError):
            gzip.GzipFile("sdnklsdnlkvsv.gz")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("sdnklsdnlkvsv.gz",)
        )

        with pytest.raises(FileNotFoundError):
            gzip.GzipFile(filename="jsgljlskjglkds.gz")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("jsgljlskjglkds.gz",)
        )

        with pytest.raises(FileNotFoundError):
            gzip.GzipFile(filename="/kjtkl.gz")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("/kjtkl.gz",)
        )


def test_gzip_open():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import gzip

        with pytest.raises(FileNotFoundError):
            gzip.open("sgsglksjks.gz")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("sgsglksjks.gz",)
        )

        with pytest.raises(FileNotFoundError):
            gzip.open(filename="qjrpoiuzu.gz")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("qjrpoiuzu.gz",)
        )



import pytest
from aikido_zen.helpers.try_parse_url_path import try_parse_url_path


def test_try_parse_url_path_nothing_found():
    assert try_parse_url_path("abc") is None


def test_try_parse_url_path_for_root():
    assert try_parse_url_path("/") == "/"


def test_try_parse_url_path_for_relative_url():
    assert try_parse_url_path("/posts") == "/posts"


def test_try_parse_url_path_for_relative_url_with_query():
    assert try_parse_url_path("/posts?abc=def") == "/posts"


def test_try_parse_url_path_for_absolute_url():
    assert try_parse_url_path("http://localhost/posts/3") == "/posts/3"


def test_try_parse_url_path_for_absolute_url_with_query():
    assert try_parse_url_path("http://localhost/posts/3?abc=def") == "/posts/3"


def test_try_parse_url_path_for_absolute_url_with_hash():
    assert try_parse_url_path("http://localhost/posts/3#abc") == "/posts/3"


def test_try_parse_url_path_for_absolute_url_with_query_and_hash():
    assert try_parse_url_path("http://localhost/posts/3?abc=def#ghi") == "/posts/3"


def test_try_parse_url_path_for_absolute_url_with_query_and_hash_no_path():
    assert try_parse_url_path("http://localhost/?abc=def#ghi") == "/"


def test_try_parse_url_path_for_absolute_url_with_query_no_path():
    assert try_parse_url_path("http://localhost?abc=def") == "/"


def test_try_parse_url_path_for_absolute_url_with_hash_no_path():
    assert try_parse_url_path("http://localhost#abc") == "/"



"""Exports get_manager_info function"""

import socket
import platform
import aikido_zen.config as config
import aikido_zen.helpers.get_machine_ip as h


def get_manager_info(connection_manager):
    """
    This returns info about the connection_manager
    """
    return {
        "dryMode": not connection_manager.block,
        "hostname": socket.gethostname(),
        "version": config.PKG_VERSION,
        "library": "firewall-python",
        "ipAddress": h.get_ip(),
        "packages": {
            pkg: details["version"]
            for pkg, details in connection_manager.packages.items()
            if "version" in details and "supported" in details and details["supported"]
        },
        "serverless": bool(connection_manager.serverless),
        "stack": list(connection_manager.packages.keys())
        + ([connection_manager.serverless] if connection_manager.serverless else []),
        "os": {"name": platform.system(), "version": platform.release()},
        "preventedPrototypePollution": False,  # Get this out of the API maybe?
        "nodeEnv": "",
        "platform": {
            "name": platform.python_implementation(),
            "version": platform.python_version(),
        },
    }



"""Export Hostnames class"""


class Hostnames:
    """Stores hostnames"""

    def __init__(self, max_entries=200):
        self.max_entries = max_entries
        self.map = {}

    def add(self, hostname, port, hits=1):
        """Add a hostname and port to the map"""
        key = get_key(hostname, port)
        if not self.map.get(key):
            self.map[key] = {"hostname": hostname, "port": port, "hits": 0}
        if len(self.map) > self.max_entries:
            # Remove the first added hostname
            first_added = next(iter(self.map))
            del self.map[first_added]
        self.map[key]["hits"] += hits

    def as_array(self):
        """Exports the contents as an array"""
        return list(self.map.values())

    def clear(self):
        """Clear the entire map"""
        self.map.clear()


def get_key(hostname, port):
    """Returns a string key"""
    return f"{hostname}:{port}"



"""Mainly exports process_read_property"""

from aikido_zen.helpers.logging import logger


def process_read_property(connection_manager, data, queue=None):
    """
    Takes in one arg : name of property on connection_manager, tries to read it.
    Meant to get config props
    """
    try:
        return connection_manager.__dict__[data]
    except KeyError:
        logger.debug(
            "CloudConnectionManager has no attribute %s, current connection_manager: %s",
            data,
            connection_manager,
        )
        return None



"""
Helper function file, see function docstring
"""

import os


def aikido_disabled_flag_active():
    """
    Checks the environment variable "AIKIDO_DISABLE"
    """
    aikido_disabled_env = os.getenv("AIKIDO_DISABLE")
    if aikido_disabled_env is not None:
        return aikido_disabled_env.lower() in ["true", "1"]
    return False



import pytest
from unittest.mock import patch
import aikido_zen.sinks.io


kind = "path_traversal"


def test_io_open():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import io

        io.open("Makefile", mode="a")
        op = "io.open"
        args1 = ("Makefile",)
        mock_run_vulnerability_scan.assert_called_once_with(
            kind=kind, op=op, args=args1
        )


def test_io_open_2():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import io

        io.open("./Makefile", mode="a")
        op = "io.open"
        args1 = ("./Makefile",)
        mock_run_vulnerability_scan.assert_called_once_with(
            kind=kind, op=op, args=args1
        )


def test_io_open_3():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import io

        io.open("/etc/hosts", mode="r")
        op = "io.open"
        args1 = ("/etc/hosts",)
        mock_run_vulnerability_scan.assert_called_once_with(
            kind=kind, op=op, args=args1
        )


def test_io_open_code():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import io

        io.open_code("Makefile")
        op = "io.open_code"
        args1 = ("Makefile",)
        mock_run_vulnerability_scan.assert_called_once_with(
            kind=kind, op=op, args=args1
        )


def test_io_open_code_2():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import io

        io.open_code("./Makefile")
        op = "io.open_code"
        args1 = ("./Makefile",)
        mock_run_vulnerability_scan.assert_called_once_with(
            kind=kind, op=op, args=args1
        )


def test_io_open_code_3():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import io

        io.open_code("/etc/hosts")
        op = "io.open_code"
        args1 = ("/etc/hosts",)
        mock_run_vulnerability_scan.assert_called_once_with(
            kind=kind, op=op, args=args1
        )



"""Exports merge_types function"""

from .is_primitive_type import only_contains_primitive_types


def merge_types(first, second):
    """Merge types into one schema if they are different."""

    # Currently we do not support merging arrays and other objects and
    # arrays/objects with primitive types
    first_primitive = only_contains_primitive_types(first["type"])
    second_primitive = only_contains_primitive_types(second["type"])
    if not first_primitive or not second_primitive:
        if first["type"] == "null":  # Prefer non-null type
            return second
        return first
    first["type"] = merge_type_arrays(first["type"], second["type"])
    return first


def merge_type_arrays(first, second):
    """Merge two lists of types into one"""
    if not isinstance(first, list):
        first = [first]

    if not isinstance(second, list):
        second = [second]
    return list(set(first).union(set(second)))  # Merging and removing duplicates



import pytest
from .route_to_key import route_to_key


def gen_route_metadata(method="GET", route="/test", url="http://localhost:5000"):
    return {"method": method, "route": route, "url": url}


def test_route_to_key_get():
    assert (
        route_to_key(gen_route_metadata(method="GET", route="/api/resource"))
        == "GET:/api/resource"
    )


def test_route_to_key_post():
    assert (
        route_to_key(gen_route_metadata(method="POST", route="/api/resource"))
        == "POST:/api/resource"
    )


def test_route_to_key_put():
    assert (
        route_to_key(gen_route_metadata(method="PUT", route="/api/resource"))
        == "PUT:/api/resource"
    )


def test_route_to_key_delete():
    assert (
        route_to_key(gen_route_metadata(method="DELETE", route="/api/resource"))
        == "DELETE:/api/resource"
    )


def test_route_to_key_with_query_params():
    assert (
        route_to_key(gen_route_metadata(method="GET", route="/api/resource?query=1"))
        == "GET:/api/resource?query=1"
    )


def test_route_to_key_with_trailing_slash():
    assert (
        route_to_key(gen_route_metadata(method="GET", route="/api/resource/"))
        == "GET:/api/resource/"
    )


def test_route_to_key_empty_path():
    assert route_to_key(gen_route_metadata(method="GET", route="")) == "GET:"



import pytest
from .normalize_asgi_headers import normalize_asgi_headers


def test_normalize_asgi_headers_basic():
    headers = [
        (b"content-type", b"text/html"),
        (b"accept-encoding", b"gzip, deflate"),
    ]
    expected = {"CONTENT_TYPE": "text/html", "ACCEPT_ENCODING": "gzip, deflate"}
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_empty():
    headers = []
    expected = {}
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_with_special_characters():
    headers = [
        (b"content-type", b"text/html"),
        (b"x-custom-header", b"some_value"),
        (b"accept-encoding", b"gzip, deflate"),
    ]
    expected = {
        "CONTENT_TYPE": "text/html",
        "X_CUSTOM_HEADER": "some_value",
        "ACCEPT_ENCODING": "gzip, deflate",
    }
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_with_dashes():
    headers = [
        (b"X-Forwarded-For", b"192.168.1.1"),
        (b"X-Request-ID", b"abc123"),
    ]
    expected = {"X_FORWARDED_FOR": "192.168.1.1", "X_REQUEST_ID": "abc123"}
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_case_insensitivity():
    headers = [
        (b"Content-Type", b"text/html"),
        (b"ACCEPT-ENCODING", b"gzip, deflate"),
    ]
    expected = {"CONTENT_TYPE": "text/html", "ACCEPT_ENCODING": "gzip, deflate"}
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_unicode():
    headers = [
        (b"content-type", b"text/html"),
        (b"custom-header", b"value"),
    ]
    expected = {"CONTENT_TYPE": "text/html", "CUSTOM_HEADER": "value"}
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_mixed_case_and_dashes():
    headers = [
        (b"Content-Type", b"text/html"),
        (b"X-Custom-Header", b"some_value"),
        (b"Accept-Encoding", b"gzip, deflate"),
    ]
    expected = {
        "CONTENT_TYPE": "text/html",
        "X_CUSTOM_HEADER": "some_value",
        "ACCEPT_ENCODING": "gzip, deflate",
    }
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_non_ascii():
    headers = [
        (b"content-type", b"text/html"),
        (b"custom-header", b"value"),
        (b"X-Header-With-Emoji", b"test"),
    ]
    expected = {
        "CONTENT_TYPE": "text/html",
        "CUSTOM_HEADER": "value",
        "X_HEADER_WITH_EMOJI": "test",
    }
    assert normalize_asgi_headers(headers) == expected


def test_normalize_asgi_headers_large_input():
    headers = [
        (f"header-{i}".encode("utf-8"), f"value-{i}".encode("utf-8"))
        for i in range(1000)
    ]
    expected = {f"HEADER_{i}": f"value-{i}" for i in range(1000)}
    assert normalize_asgi_headers(headers) == expected



from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import before, on_import, patch_function
from aikido_zen.vulnerabilities import run_vulnerability_scan


@before
def _execute(func, instance, args, kwargs):
    query = get_argument(args, kwargs, 0, "query")

    op = "clickhouse_driver.Client.execute"
    register_call(op, "sql_op")

    run_vulnerability_scan("sql_injection", op, args=(query, "clickhouse"))


@on_import("clickhouse_driver", package="clickhouse_driver")
def patch(m):
    """
    patching module clickhouse_driver
    - patches clickhouse_driver.Client.execute
    - patches clickhouse_driver.Client.execute_iter
    - patches clickhouse_driver.Client.execute_with_progress
    """
    patch_function(m, "Client.execute", _execute)
    patch_function(m, "Client.execute_iter", _execute)
    patch_function(m, "Client.execute_with_progress", _execute)



import pytest
from unittest.mock import MagicMock, patch
import aikido_zen.context as ctx
from .extract_data_from_xml_body import (
    extract_data_from_xml_body,
)  # Replace 'your_module' with the actual module name


@pytest.fixture
def mock_context():
    mock_ctx = MagicMock()
    mock_ctx.body = "valid_input"
    mock_ctx.xml = {}  # Initialize with an empty dictionary
    return mock_ctx


def test_does_not_crash_when_context_none(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=None):
        user_input = "valid_input"
        root_element = [
            {"attr1": "value1", "attr2": "value2"},
            {"attr1": "value3", "attr3": "value4"},
        ]

        extract_data_from_xml_body(user_input, root_element)


def test_extract_data_from_xml_body_valid_input(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        user_input = "valid_input"
        root_element = [
            {"attr1": "value1", "attr2": "value2"},
            {"attr1": "value3", "attr3": "value4"},
        ]

        extract_data_from_xml_body(user_input, root_element)

        assert mock_context.xml == {
            "attr1": {"value1", "value3"},
            "attr2": {"value2"},
            "attr3": {"value4"},
        }


def test_extract_data_from_xml_body_invalid_user_input(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        user_input = "invalid_input"
        root_element = [{"attr1": "value1"}]

        extract_data_from_xml_body(user_input, root_element)

        assert mock_context.xml == {}


def test_extract_data_from_xml_body_empty_root_element(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        user_input = "valid_input"
        root_element = []

        extract_data_from_xml_body(user_input, root_element)

        assert mock_context.xml == {}


def test_extract_data_from_xml_body_non_string_context_body(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        mock_context.body = 123  # Set body to a non-string value
        user_input = "valid_input"
        root_element = [{"attr1": "value1"}]

        extract_data_from_xml_body(user_input, root_element)

        assert mock_context.xml == {}


def test_extract_data_from_xml_body_multiple_calls(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        user_input = "valid_input"
        root_element1 = [{"attr1": "value1"}]
        root_element2 = [{"attr1": "value2"}]

        extract_data_from_xml_body(user_input, root_element1)
        extract_data_from_xml_body(user_input, root_element2)

        assert mock_context.xml == {"attr1": {"value1", "value2"}}


def test_extract_data_from_xml_body_duplicate_attributes(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        user_input = "valid_input"
        root_element = [
            {"attr1": "value1"},
            {"attr1": "value1"},  # Duplicate
            {"attr2": "value2"},
        ]

        extract_data_from_xml_body(user_input, root_element)

        assert mock_context.xml == {"attr1": {"value1"}, "attr2": {"value2"}}


def test_extract_data_from_xml_body_no_attributes(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        user_input = "valid_input"
        root_element = [{}]

        extract_data_from_xml_body(user_input, root_element)

        assert mock_context.xml == {}


def test_extract_data_from_xml_body_context_set_as_current(mock_context):
    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        user_input = "valid_input"
        root_element = [{"attr1": "value1"}]

        extract_data_from_xml_body(user_input, root_element)

        mock_context.set_as_current_context.assert_called_once()



import pytest
from .protection_forced_off import protection_forced_off


def sample_route_metadata(route=None, url=None):
    # This is a mock function to simulate the behavior of the actual sample_route_metadata function.
    return {"route": route, "url": url, "method": "POST"}


def test_protection_forced_off_with_protection_off():
    route_metadata = sample_route_metadata(
        url="http://example.com/posts/3", route="/posts/:number"
    )
    endpoints = [
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": True,  # This should return True
        },
    ]
    assert protection_forced_off(route_metadata, endpoints) is True


def test_protection_forced_off_with_protection_on():
    route_metadata = sample_route_metadata(
        url="http://example.com/posts/3", route="/posts/:number"
    )
    endpoints = [
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,  # This should return False
        },
    ]
    assert protection_forced_off(route_metadata, endpoints) is False


def test_protection_forced_off_with_no_matches():
    route_metadata = sample_route_metadata(
        url="http://example.com/unknown", route="/unknown"
    )
    endpoints = [
        {
            "method": "GET",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]
    assert protection_forced_off(route_metadata, endpoints) is False


def test_protection_forced_off_with_empty_endpoints():
    route_metadata = sample_route_metadata(
        url="http://example.com/posts/3", route="/posts/:number"
    )
    endpoints = []
    assert protection_forced_off(route_metadata, endpoints) is False


def test_protection_forced_off_with_multiple_endpoints():
    route_metadata = sample_route_metadata(
        url="http://example.com/posts/3", route="/posts/:number"
    )
    endpoints = [
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": True,
        },
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]
    assert protection_forced_off(route_metadata, endpoints) is True


def test_protection_forced_off_with_multiple_endpoints2():
    route_metadata = sample_route_metadata(
        url="http://example.com/posts/3", route="/posts/:number"
    )
    endpoints = [
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": True,
        },
    ]
    assert protection_forced_off(route_metadata, endpoints) is True


def test_protection_forced_off_with_multiple_endpoints3():
    route_metadata = sample_route_metadata(
        url="http://example.com/posts/3", route="/posts/:number"
    )
    endpoints = [
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]
    assert protection_forced_off(route_metadata, endpoints) is False



from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.logging import logger
from aikido_zen.context import Context
from aikido_zen.context import get_current_context
import aikido_zen.sources.functions.request_handler as funcs
from aikido_zen.sinks import (
    on_import,
    patch_function,
    after,
    before_modify_return,
    before,
)


@before_modify_return
def _full_dispatch_request_before(func, instance, args, kwargs):
    from flask.globals import request_ctx
    from flask import Response

    req = request_ctx.request
    extract_cookies_from_flask_request_and_save_data(req)
    extract_form_data_from_flask_request_and_save_data(req)
    extract_view_args_from_flask_request_and_save_data(req)

    pre_response = funcs.request_handler(stage="pre_response")
    if not pre_response:
        return None
    # This happens when a route is rate limited, a user blocked, etc...
    return Response(pre_response[0], status=pre_response[1], mimetype="text/plain")


@after
def _full_dispatch_request_after(func, instance, args, kwargs, return_value):
    if not hasattr(return_value, "status_code"):
        return
    funcs.request_handler(stage="post_response", status_code=return_value.status_code)


def extract_view_args_from_flask_request_and_save_data(req):
    """Extract view args from flask request"""
    context = get_current_context()

    try:
        if getattr(req, "view_args"):
            context.route_params = dict(req.view_args)
            context.set_as_current_context()
    except Exception as e:
        logger.debug("Exception occurred whilst extracting flask view args data: %s", e)


def extract_form_data_from_flask_request_and_save_data(req):
    """Extract form data from flask request"""
    context = get_current_context()
    try:
        if context:
            if req.form:
                context.set_body(req.form)
            else:
                context.set_body(req.data.decode("utf-8"))
            context.set_as_current_context()
    except Exception as e:
        logger.debug("Exception occurred whilst extracting flask body data: %s", e)


def extract_cookies_from_flask_request_and_save_data(req):
    """Extract cookies from flask request"""
    context = get_current_context()
    try:
        context.cookies = req.cookies.to_dict()
        context.set_as_current_context()
    except Exception as e:
        logger.debug("Exception occurred whilst extracting flask cookie data: %s", e)


@before
def _call(func, instance, args, kwargs):
    environ = get_argument(args, kwargs, 0, "environ")
    context1 = Context(req=environ, source="flask")
    context1.set_as_current_context()
    funcs.request_handler(stage="init")


@on_import("flask.app", "flask", version_requirement="2.2.4")
def patch(m):
    """
    patching module flask.appimport

    - patches: Flask.__call__ (context parsing/initial stage)
    - patches: Flask.full_dispatch_request
    **Why?** full_dispatch_request gets called in the WSGI app. It gets called after all middleware. request_ctx is
    available, so we can extract data from it. It returns a response, so we can send status codes and error messages.
    (github src: https://github.com/pallets/flask/blob/bc143499cf1137a271a7cf75bdd3e16e43ede2f0/src/flask/app.py#L1529)
    """
    patch_function(m, "Flask.__call__", _call)
    patch_function(m, "Flask.full_dispatch_request", _full_dispatch_request_before)
    patch_function(m, "Flask.full_dispatch_request", _full_dispatch_request_after)



import pytest
from .operations import Operations


@pytest.fixture
def operations():
    return Operations()


def test_ensure_operation(operations):
    operation = "test_op"
    kind = "sql_op"
    operations.ensure_operation(operation, kind)
    assert operation in operations
    assert operations[operation]["kind"] == kind
    assert operations[operation]["total"] == 0
    assert operations[operation]["attacksDetected"]["total"] == 0
    assert operations[operation]["attacksDetected"]["blocked"] == 0


def test_ensure_operation_unsupported_kind(operations):
    operation = "test_op"
    kind = "unsupported_kind"
    with pytest.raises(Exception):
        operations.ensure_operation(operation, kind)


def test_register_call(operations):
    operation = "test_op"
    kind = "sql_op"
    operations.register_call(operation, kind)
    assert operation in operations
    assert operations[operation]["total"] == 1


def test_on_detected_attack(operations):
    operation = "test_op"
    kind = "ai_op"
    operations.ensure_operation(operation, kind)
    operations.on_detected_attack(blocked=True, operation=operation)
    assert operations[operation]["attacksDetected"]["total"] == 1
    assert operations[operation]["attacksDetected"]["blocked"] == 1


def test_on_detected_attack_not_blocked(operations):
    operation = "test_op"
    kind = "sql_op"
    operations.ensure_operation(operation, kind)
    operations.on_detected_attack(blocked=False, operation=operation)
    assert operations[operation]["attacksDetected"]["total"] == 1
    assert operations[operation]["attacksDetected"]["blocked"] == 0


def test_on_detected_attack_unknown_operation(operations):
    operation = "unknown_op"
    operations.on_detected_attack(blocked=True, operation=operation)
    assert operation not in operations


def test_register_call_and_on_detected_attack(operations):
    operation = "test_op"
    kind = "sql_op"
    operations.register_call(operation, kind)
    operations.on_detected_attack(blocked=True, operation=operation)
    assert operations[operation]["total"] == 1
    assert operations[operation]["attacksDetected"]["total"] == 1
    assert operations[operation]["attacksDetected"]["blocked"] == 1



"""
Helper function file, see function docstring
"""

import os


def check_env_for_blocking():
    """
    Checks the environment variable "AIKIDO_BLOCK"
    """
    # Set log level
    aikido_blocking_env = os.getenv("AIKIDO_BLOCK")
    if aikido_blocking_env is None:
        aikido_blocking_env = os.getenv(
            "AIKIDO_BLOCKING"
        )  # Fallback to AIKIDO_BLOCKING.
    if aikido_blocking_env is not None:
        return aikido_blocking_env.lower() in ["true", "1"]
    return False



import pytest
import time
from .rate_limiter import RateLimiter


@pytest.fixture
def rate_limiter():
    max_items = 10
    time_to_live_in_ms = 1000  # 1 second TTL
    return RateLimiter(max_items, time_to_live_in_ms)


def test_allow_requests_within_limit(rate_limiter):
    key = "user1"
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, 5000000, 5
        ), f"Request {i + 1} should be allowed"


def test_deny_requests_exceeding_limit(rate_limiter):
    key = "user2"
    for i in range(5):
        rate_limiter.is_allowed(key, 500, 5)
    assert not rate_limiter.is_allowed(key, 500, 5), "Request 6 should not be allowed"


def test_clear_old_entries(rate_limiter):
    key = "user3"
    for i in range(5):
        rate_limiter.is_allowed(key, 500, 5)

    time.sleep(0.6)  # Sleep to allow old entries to be cleared

    assert rate_limiter.is_allowed(
        key, 500, 5
    ), "New request should be allowed after clearing old entries"


def test_multiple_keys(rate_limiter):
    key1 = "user4"
    key2 = "user5"

    for i in range(5):
        assert rate_limiter.is_allowed(
            key1, 500, 5
        ), f"Request {i + 1} for key1 should be allowed"
    for i in range(5):
        assert rate_limiter.is_allowed(
            key2, 500, 5
        ), f"Request {i + 1} for key2 should be allowed"

    assert not rate_limiter.is_allowed(
        key1, 500, 5
    ), "Request 6 for key1 should not be allowed"
    assert not rate_limiter.is_allowed(
        key2, 500, 5
    ), "Request 6 for key2 should not be allowed"


def test_ttl_expiration(rate_limiter):
    key = "user6"
    for i in range(5):
        rate_limiter.is_allowed(key, 500, 5)

    time.sleep(1.1)  # Sleep to allow the TTL to expire

    assert rate_limiter.is_allowed(key, 500, 5), "Request after TTL should be allowed"


def test_allow_requests_exactly_at_limit(rate_limiter):
    key = "user7"
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, 500, 5
        ), f"Request {i + 1} should be allowed"
    assert not rate_limiter.is_allowed(key, 500, 5), "Request 6 should not be allowed"


def test_allow_requests_after_clearing_old_entries(rate_limiter):
    key = "user8"
    for i in range(5):
        rate_limiter.is_allowed(key, 500, 5)

    time.sleep(0.6)  # Sleep to allow old entries to be cleared

    assert rate_limiter.is_allowed(
        key, 500, 5
    ), "New request should be allowed after clearing old entries"


def test_multiple_rapid_requests(rate_limiter):
    key = "user9"
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, 500, 5
        ), f"Request {i + 1} should be allowed"

    time.sleep(0.1)  # Sleep for 100 ms

    assert not rate_limiter.is_allowed(
        key, 500, 5
    ), "Request after rapid requests should not be allowed"


def test_reset_after_ttl(rate_limiter):
    key = "user10"
    for i in range(5):
        rate_limiter.is_allowed(key, 500, 5)

    time.sleep(1.1)  # Sleep to allow the TTL to expire

    assert rate_limiter.is_allowed(key, 500, 5), "Request after TTL should be allowed"


def test_different_window_sizes(rate_limiter):
    key = "user11"
    different_window_size = 1000  # 1 second window
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, different_window_size, 5
        ), f"Request {i + 1} should be allowed"
    assert not rate_limiter.is_allowed(
        key, different_window_size, 5
    ), "Request 6 should not be allowed"


def test_sliding_window_with_intermittent_requests(rate_limiter):
    key = "user14"

    # Allow 5 requests in a 1-second window
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, 500, 5
        ), f"Request {i + 1} should be allowed"
        time.sleep(0.1)  # Sleep 100 ms between requests

    # Sleep for 600 ms to allow the first requests to slide out of the window
    time.sleep(0.6)

    # Now we should be able to make a new request
    assert rate_limiter.is_allowed(
        key, 500, 5
    ), "New request should be allowed after sliding"


def test_sliding_window_edge_case(rate_limiter):
    key = "user15"

    # Allow 5 requests in a 1-second window
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, 500, 5
        ), f"Request {i + 1} should be allowed"

    # Sleep for 500 ms to simulate time passing
    time.sleep(0.6)

    # The next request should still be allowed as the window is sliding
    assert rate_limiter.is_allowed(
        key, 500, 5
    ), "Next request should be allowed as window slides"

    # Sleep for another 500 ms to allow the first batch to expire
    time.sleep(0.6)

    # Now we should be able to make a new request
    assert rate_limiter.is_allowed(
        key, 500, 5
    ), "New request should be allowed after first batch expires"


def test_sliding_window_with_burst_requests(rate_limiter):
    key = "user16"
    window_size_ms = 500

    # Allow 5 requests in a 1-second window
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, window_size_ms, 5
        ), f"Request {i + 1} should be allowed"

    # Sleep for 250 ms to simulate time passing
    time.sleep((window_size_ms / 2) / 1000)

    # Add 3 more requests (should be denied)
    assert not rate_limiter.is_allowed(
        key, window_size_ms, 5
    ), "Request should not be allowed"
    assert not rate_limiter.is_allowed(
        key, window_size_ms, 5
    ), "Request should not be allowed"
    assert not rate_limiter.is_allowed(
        key, window_size_ms, 5
    ), "Request should not be allowed"

    time.sleep(
        (window_size_ms / 2 + 50) / 1000
    )  # 50ms buffer, first requests are still in window

    # Make a burst of requests (should be allowed)
    for i in range(2):
        assert rate_limiter.is_allowed(
            key, window_size_ms, 5
        ), f"Burst request {i + 1} should be allowed"
    assert not rate_limiter.is_allowed(
        key, window_size_ms, 5
    ), "Burst request should not be allowed after limit"

    # Sleep for 500 ms to allow all batches to slide out of window
    time.sleep((window_size_ms + 100) / 1000)

    # Now we should be able to make a new request
    assert rate_limiter.is_allowed(
        key, window_size_ms, 5
    ), "New request should be allowed after all batches slide out"


def test_sliding_window_with_delayed_requests(rate_limiter):
    key = "user17"

    # Allow 5 requests in a 1-second window
    for i in range(5):
        assert rate_limiter.is_allowed(
            key, 500, 5
        ), f"Request {i + 1} should be allowed"
        time.sleep(0.1)  # Sleep 100 ms between requests

    # Sleep for 600 ms to allow the first requests to slide out of the window
    time.sleep(0.6)

    # Now we should be able to make a new request
    assert rate_limiter.is_allowed(
        key, 500, 5
    ), "New request should be allowed after sliding"



"""
Helper function file, see funtion definition
"""

from collections.abc import Mapping


def is_mapping(o):
    """
    Checks if the object is a plain object,
    i.e. an instance of Mapping in Python
    """
    return isinstance(o, Mapping)



"""Exports the send_heartbeat function"""

from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms


def send_heartbeat(connection_manager):
    """
    This will send a heartbeat to the server
    """
    if not connection_manager.token:
        return
    logger.debug("Aikido CloudConnectionManager : Sending out heartbeat")
    stats = connection_manager.statistics.get_record()
    users = connection_manager.users.as_array()
    routes = list(connection_manager.routes)
    outgoing_domains = connection_manager.hostnames.as_array()
    ai_stats = connection_manager.ai_stats.get_stats()

    connection_manager.statistics.clear()
    connection_manager.users.clear()
    connection_manager.routes.clear()
    connection_manager.hostnames.clear()
    connection_manager.ai_stats.clear()
    res = connection_manager.api.report(
        connection_manager.token,
        {
            "type": "heartbeat",
            "time": get_unixtime_ms(),
            "agent": connection_manager.get_manager_info(),
            "stats": stats,
            "ai": ai_stats,
            "hostnames": outgoing_domains,
            "routes": routes,
            "users": users,
            "middlewareInstalled": connection_manager.middleware_installed,
        },
        connection_manager.timeout_in_sec,
    )
    connection_manager.update_service_config(res)



"""Mainly exports merge_auth_types"""


def merge_auth_types(existing, new_auth):
    """
    Merge two auth type lists into one, without duplicates. (See get_auth_types)
    Can return None if both parameters are not a list.
    """
    if not isinstance(new_auth, list) or len(new_auth) == 0:
        return existing

    if not isinstance(existing, list) or len(existing) == 0:
        return new_auth

    result = existing.copy()

    for auth in new_auth:
        if not any(is_equal_api_auth_type(a, auth) for a in result):
            result.append(auth)

    return result


def is_equal_api_auth_type(a, b):
    """
    Compare two APIAuthType objects for equality.
    """
    return (
        a.get("type") == b.get("type")
        and a.get("in") == b.get("in")
        and a.get("name") == b.get("name")
        and a.get("scheme") == b.get("scheme")
    )



"""
Sink module for `os`, wrapping os.system
"""

import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import patch_function, before, on_import


@before
def _system(func, instance, args, kwargs):
    command = get_argument(args, kwargs, 0, "command")
    if not isinstance(command, str):
        return

    op = "os.system"
    register_call(op, "exec_op")

    vulns.run_vulnerability_scan(kind="shell_injection", op=op, args=(command,))


@on_import("os")
def patch(m):
    """
    patching os module
    - patches os.system for shell injection
    - does not patch: os.popen -> uses subprocess.Popen
    - does not patch: os.execl, os.execle, os.execlp, ... -> only argument injection
    """
    patch_function(m, "system", _system)



import pytest
from unittest.mock import patch, MagicMock
from aikido_zen.thread.thread_cache import get_cache, ThreadCache
from .request_handler import request_handler
from ...background_process.service_config import ServiceConfig
from ...context import Context, current_context


@pytest.fixture
def mock_context():
    """Fixture to create a mock context."""
    context = MagicMock()
    context.route = "/test/route"
    context.method = "GET"
    context.get_route_metadata.return_value = {
        "route": "/test/route",
        "method": "GET",
        "url": "http://localhost:8080/test/route",
    }
    context.body = {}
    return context


@pytest.fixture(autouse=True)
def run_around_tests():
    get_cache().reset()
    current_context.set(None)
    yield
    get_cache().reset()
    current_context.set(None)


def test_post_response_useful_route(mock_context):
    """Test post_response when the route is useful."""

    cache = get_cache()  # Creates a new cache
    assert cache.routes.routes == {}
    for i in range(25):
        with patch("aikido_zen.context.get_current_context", return_value=mock_context):
            mock_context.body[str(i + 1)] = i + 1
            request_handler("post_response", status_code=200)

    # Check that the route was initialized and updated
    route = cache.routes.routes.get("GET:/test/route")
    assert route["hits"] == route["hits_delta_since_sync"] == 25
    assert route["method"] == "GET"
    assert route["path"] == "/test/route"

    # Test only updates on first 20 hits
    body_props = route["apispec"]["body"]["schema"]["properties"]
    assert body_props == {
        "1": {"type": "number"},
        "2": {"type": "number", "optional": True},
        "3": {"type": "number", "optional": True},
        "4": {"type": "number", "optional": True},
        "5": {"type": "number", "optional": True},
        "6": {"type": "number", "optional": True},
        "7": {"type": "number", "optional": True},
        "8": {"type": "number", "optional": True},
        "9": {"type": "number", "optional": True},
        "10": {"type": "number", "optional": True},
        "11": {"type": "number", "optional": True},
        "12": {"type": "number", "optional": True},
        "13": {"type": "number", "optional": True},
        "14": {"type": "number", "optional": True},
        "15": {"type": "number", "optional": True},
        "16": {"type": "number", "optional": True},
        "17": {"type": "number", "optional": True},
        "18": {"type": "number", "optional": True},
        "19": {"type": "number", "optional": True},
        "20": {"type": "number", "optional": True},
    }


@patch("aikido_zen.background_process.get_comms")
def test_post_response_not_useful_route(mock_get_comms, mock_context):
    """Test post_response when the route is not useful."""
    comms = MagicMock()
    mock_get_comms.return_value = comms

    cache = ThreadCache()  # Creates a new cache
    assert cache.routes.routes == {}

    with patch("aikido_zen.context.get_current_context", return_value=mock_context):
        request_handler("post_response", status_code=500)

    assert cache.routes.routes == {}
    comms.send_data_to_bg_process.assert_not_called()


@patch("aikido_zen.background_process.get_comms")
def test_post_response_no_context(mock_get_comms):
    """Test post_response when there is no context."""
    comms = MagicMock()
    mock_get_comms.return_value = comms

    cache = ThreadCache()  # Creates a new cache
    assert cache.routes.routes == {}

    # Simulate no context
    with patch("aikido_zen.context.get_current_context", return_value=None):
        result = request_handler("post_response", status_code=200)

    assert cache.routes.routes == {}
    comms.send_data_to_bg_process.assert_not_called()


# Test firewall lists
def set_context(remote_address, user_agent=""):
    Context(
        context_obj={
            "remote_address": remote_address,
            "method": "POST",
            "url": "http://localhost:4000",
            "query": {"abc": "def"},
            "headers": {"USER_AGENT": user_agent},
            "body": None,
            "cookies": {},
            "source": "flask",
            "route": "/posts/:number",
            "user": None,
            "executed_middleware": False,
        }
    ).set_as_current_context()


def create_service_config(blocked_ips=None):
    config = ServiceConfig(
        endpoints=[
            {
                "method": "POST",
                "route": "/posts/:number",
                "graphql": False,
                "allowedIPAddresses": ["1.1.1.1", "2.2.2.2", "3.3.3.3"],
            }
        ],
        last_updated_at=None,
        blocked_uids=set(),
        bypassed_ips=[],
        received_any_stats=False,
    )
    if blocked_ips:
        config.set_blocked_ips(blocked_ips)
    get_cache().config = config
    return config


def test_blocked_ip():
    # Arrange
    set_context("192.168.1.1")
    blocked_ips = [
        {"source": "test", "description": "Blocked for testing", "ips": ["192.168.1.1"]}
    ]
    config = create_service_config(blocked_ips)
    config.endpoints[0]["allowedIPAddresses"] = []  # Clear allowed ips for endpoint

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is blocked due to Blocked for testing (Your IP: 192.168.1.1)",
        403,
    )


def test_allowed_ip():
    # Arrange
    set_context("1.1.1.1")
    blocked_ips = [
        {"source": "test", "description": "Blocked for testing", "ips": ["192.168.1.1"]}
    ]
    create_service_config(blocked_ips)

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result is None


def test_invalid_context():
    # Arrange
    create_service_config()

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result is None


def test_not_allowed_ip():
    # Arrange
    set_context("192.168.1.3")
    blocked_ips = [
        {"source": "test", "description": "Blocked for testing", "ips": ["192.168.1.1"]}
    ]
    create_service_config(blocked_ips)

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is not allowed to access this resource. (Your IP: 192.168.1.3)",
        403,
    )


def test_bypassed_ip():
    # Arrange
    set_context("1.1.1.1")  # This IP is in the allowed list
    blocked_ips = [
        {"source": "test", "description": "Blocked for testing", "ips": ["192.168.1.1"]}
    ]
    config = create_service_config(blocked_ips)
    config.bypassed_ips.add("1.1.1.1")

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result is None  # Should be allowed since it's in the bypass list


def test_ip_allowed_by_endpoint():
    # Arrange
    set_context("2.2.2.2")  # This IP is in the allowed list
    blocked_ips = [
        {"source": "test", "description": "Blocked for testing", "ips": ["192.168.1.1"]}
    ]
    config = create_service_config(blocked_ips)

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result is None  # Should be allowed since it's in the allowed list


def test_ip_not_allowed_by_endpoint():
    # Arrange
    set_context("4.4.4.4")  # Not in the allowed list
    blocked_ips = [
        {"source": "test", "description": "Blocked for testing", "ips": ["192.168.1.1"]}
    ]
    config = create_service_config(blocked_ips)

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is not allowed to access this resource. (Your IP: 4.4.4.4)",
        403,
    )


def test_first_checks_resource_blocked():
    # Arrange
    set_context("192.168.1.1")  # This IP is blocked
    blocked_ips = [
        {
            "source": "test",
            "description": "Blocked for testing",
            "ips": ["192.168.1.1", "192.168.1.2"],
        }
    ]
    create_service_config(blocked_ips)

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is not allowed to access this resource. (Your IP: 192.168.1.1)",
        403,
    )


def test_allowed_for_endpoint_but_blocked():
    # Arrange
    set_context("1.1.1.1")  # This IP is blocked
    blocked_ips = [
        {
            "source": "test",
            "description": "Blocked for testing",
            "ips": ["192.168.1.1", "192.168.1.2", "1.1.1.0/24"],
        }
    ]
    create_service_config(blocked_ips)

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is blocked due to Blocked for testing (Your IP: 1.1.1.1)",
        403,
    )


def test_allowed_for_endpoint_but_not_in_allowlist_and_blocked():
    # Arrange
    set_context("1.1.1.1")  # This IP is blocked
    blocked_ips = [
        {
            "source": "test",
            "description": "Blocked for testing",
            "ips": ["192.168.1.1", "192.168.1.2", "1.1.1.0/24"],
        }
    ]
    config = create_service_config(blocked_ips)
    config.set_allowed_ips(
        [
            {
                "source": "test",
                "description": "Allowed ranges",
                "ips": ["4.4.4.0/24"],
            }
        ]
    )

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is not allowed. (Your IP: 1.1.1.1)",
        403,
    )


def test_allowed_for_endpoint_but_not_in_allowlist_not_blocked():
    # Arrange
    set_context("1.1.1.1")  # This IP is blocked
    config = create_service_config()
    config.set_allowed_ips(
        [
            {
                "source": "test",
                "description": "Allowed ranges",
                "ips": ["4.4.4.0/24"],
            }
        ]
    )

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is not allowed. (Your IP: 1.1.1.1)",
        403,
    )


def test_allowed_for_endpoint_and_in_allowlist():
    # Arrange
    set_context("1.1.1.1")  # This IP is blocked
    config = create_service_config()
    config.set_allowed_ips(
        [
            {
                "source": "test",
                "description": "Allowed ranges",
                "ips": ["1.1.1.0/24"],
            }
        ]
    )

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result is None


def test_allowed_for_endpoint_in_allowlist_but_blocked():
    # Arrange
    set_context("1.1.1.1")  # This IP is blocked
    blocked_ips = [
        {
            "source": "test",
            "description": "Blocked for testing",
            "ips": ["192.168.1.1", "192.168.1.2", "1.1.1.0/24"],
        }
    ]
    config = create_service_config(blocked_ips)
    config.set_allowed_ips(
        [
            {
                "source": "test",
                "description": "Allowed ranges",
                "ips": ["1.1.1.0/24"],
            }
        ]
    )

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "Your IP address is blocked due to Blocked for testing (Your IP: 1.1.1.1)",
        403,
    )


def test_allowed_for_endpoint_and_in_allowlist_but_is_bot():
    # Arrange
    set_context("1.1.1.1", "Test_be bot")
    config = create_service_config()
    config.set_blocked_user_agents("test_ua|test_be")

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "You are not allowed to access this resource because you have been identified as a bot.",
        403,
    )


def test_allowed_for_endpoint_and_in_allowlist_but_is_bot_2():
    # Arrange
    set_context(
        "1.1.1.1",
        "Mozilla/5.0 (compatible; Bytespider/1.0; +http://bytespider.com/bot.html)",
    )
    config = create_service_config()
    config.set_blocked_user_agents(
        "AI2Bot|Applebot-Extended|Bytespider|CCBot|ClaudeBot|cohere-training-data-crawler|Diffbot|Google-Extended|GPTBot|Kangaroo Bot|meta-externalagent|anthropic-ai|omgili|PanguBot|Webzio-Extended|Timpibot|img2dataset|ImagesiftBot|archive.org_bot"
    )

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result == (
        "You are not allowed to access this resource because you have been identified as a bot.",
        403,
    )

    set_context("1.1.1.1", "Mozilla/5.0 (compatible; GPTBot/1.0;")

    result = request_handler("pre_response")

    # Assert
    assert result == (
        "You are not allowed to access this resource because you have been identified as a bot.",
        403,
    )


def test_allowed_for_endpoint_and_in_allowlist_bots_are_set_but_is_not_bot():
    # Arrange
    set_context("1.1.1.1", "Test_u4 bot")
    config = create_service_config()
    config.set_blocked_user_agents("test_ua|test_be")

    # Act
    result = request_handler("pre_response")

    # Assert
    assert result is None



from aikido_zen.helpers.is_localhost_ip import is_localhost_ip
from aikido_zen.vulnerabilities.ssrf.is_private_ip import is_private_ip


def is_ip_allowed_by_allowlist(service_config, ip):
    """
    This function was separated from the ServiceConfig object, because otherwise you would have a circular dep.
    """
    if not service_config.allowed_ips or len(service_config.allowed_ips) < 1:
        return True
    # Always allow access from local IP addresses
    if is_private_ip(ip):
        return True

    for entry in service_config.allowed_ips:
        if entry["iplist"].matches(ip):
            # If the IP matches one of the lists the IP is allowed :
            return True
    return False



"""Usefull export : get_auth_types"""

from aikido_zen.helpers.is_http_auth_scheme import is_http_auth_scheme

common_api_key_header_names = [
    "x-api-key",
    "api-key",
    "apikey",
    "x-token",
    "token",
]

# Common auth cookie names
common_auth_cookie_names = [
    "auth",
    "session",
    "jwt",
    "token",
    "sid",
    "connect.sid",
    "auth_token",
    "access_token",
    "refresh_token",
] + common_api_key_header_names


def get_auth_types(context):
    """Get the authentication type of the API request."""
    if not isinstance(context.headers, dict):
        return None

    result = []

    # Check the Authorization header
    auth_header = context.headers.get("AUTHORIZATION")
    if isinstance(auth_header, str):
        auth_header_type = get_authorization_header_type(auth_header)
        if auth_header_type:
            result.append(auth_header_type)

    # Check for type apiKey in headers and cookies
    result.extend(find_api_keys(context))

    return result if result else None


def get_authorization_header_type(auth_header):
    """Get the authentication type from the Authorization header."""
    if not auth_header:
        return None
    if " " in auth_header:
        type_, value = auth_header.split(" ", 1)

        if isinstance(type_, str) and isinstance(value, str):
            if is_http_auth_scheme(type_):
                return {"type": "http", "scheme": type_.lower()}

    # Default to apiKey if the auth type is not recognized
    return {"type": "apiKey", "in": "header", "name": "Authorization"}


def find_api_keys(context):
    """Search for API keys in headers and cookies."""
    result = []

    for header in common_api_key_header_names:
        if normalize_header_key(header) in context.headers:
            result.append({"type": "apiKey", "in": "header", "name": header})

    if isinstance(context.cookies, dict):
        relevant_cookies = [
            cookie
            for cookie in context.cookies
            if cookie.lower() in common_auth_cookie_names
        ]
        for cookie in relevant_cookies:
            result.append({"type": "apiKey", "in": "cookie", "name": cookie})

    return result


def normalize_header_key(header_key):
    """Normalizes the header keys"""
    return header_key.replace("-", "_").upper()



import pytest

from .build_url_from_asgi import build_url_from_asgi


def test_build_url_from_asgi_http():
    scope = {
        "scheme": "http",
        "server": ("localhost", 8000),
        "root_path": "",
        "path": "/api/v1/resource",
    }
    expected = "http://localhost:8000/api/v1/resource"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_https():
    scope = {
        "scheme": "https",
        "server": ("example.com", 443),
        "root_path": "",
        "path": "/secure/resource",
    }
    expected = "https://example.com:443/secure/resource"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_root_path():
    scope = {
        "scheme": "http",
        "server": ("localhost", 8000),
        "root_path": "/api",
        "path": "/api/v1/resource",
    }
    expected = "http://localhost:8000/v1/resource"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_empty_path():
    scope = {
        "scheme": "http",
        "server": ("localhost", 8000),
        "root_path": "",
        "path": "",
    }
    expected = "http://localhost:8000"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_without_server():
    scope = {
        "scheme": "http",
        "server": None,
        "root_path": "",
        "path": "/api/v1/resource",
    }
    expected = None
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_no_root_path():
    scope = {
        "scheme": "http",
        "server": ("localhost", 8000),
        "path": "/api/v1/resource",
    }
    expected = "http://localhost:8000/api/v1/resource"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_port_80():
    scope = {
        "scheme": "http",
        "server": ("localhost", 80),
        "root_path": "",
        "path": "/",
    }
    expected = "http://localhost:80/"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_port_443():
    scope = {
        "scheme": "https",
        "server": ("localhost", 443),
        "root_path": "",
        "path": "/",
    }
    expected = "https://localhost:443/"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_trailing_slash_in_root_path():
    scope = {
        "scheme": "http",
        "server": ("localhost", 8000),
        "root_path": "/api",
        "path": "/api/v1/resource",
    }
    expected = "http://localhost:8000/v1/resource"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_multiple_slashes():
    scope = {
        "scheme": "http",
        "server": ("localhost", 8000),
        "root_path": "/api/",
        "path": "/api//v1/resource",
    }
    expected = "http://localhost:8000/v1/resource"
    assert build_url_from_asgi(scope) == expected


def test_build_url_from_asgi_with_query_string():
    scope = {
        "scheme": "http",
        "server": ("localhost", 8000),
        "root_path": "",
        "path": "/api/v1/resource?query=1",
    }
    expected = "http://localhost:8000/api/v1/resource?query=1"
    assert build_url_from_asgi(scope) == expected



"""
Helper function file, see function docstrin
"""

import base64
import json


def try_decode_as_jwt(jwt):
    """
    This will try to decode a string as a JWT
    """
    if "." not in jwt:
        return (False, None)

    parts = jwt.split(".")

    if len(parts) != 3:
        return (False, None)

    try:
        base64_decoded = base64.urlsafe_b64decode(str.encode(parts[1]) + b"==")
        jwt = json.loads(base64_decoded.decode("utf-8"))
        return (True, jwt)
    except Exception:
        return (False, None)



"""`Django` source module"""

from ..functions.request_handler import request_handler
from .run_init_stage import run_init_stage
from .pre_response_middleware import pre_response_middleware
from ...helpers.get_argument import get_argument
from ...sinks import on_import, patch_function, before, after


@before
def _get_response_before(func, instance, args, kwargs):
    request = get_argument(args, kwargs, 0, "request")

    run_init_stage(request)

    if pre_response_middleware not in instance._view_middleware:
        # The rate limiting middleware needs to be last in the chain.
        instance._view_middleware += [pre_response_middleware]


@after
def _get_response_after(func, instance, args, kwargs, return_value):
    if hasattr(return_value, "status_code"):
        request_handler(stage="post_response", status_code=return_value.status_code)


@on_import("django.core.handlers.base", "django")
def patch(m):
    """
    Patch for _get_response (Synchronous/WSGI)
    - before: Parse body, create context & add middleware to run before a response
    - after: Check respone code to see if route should be analyzed
    # https://github.com/django/django/blob/5865ff5adcf64da03d306dc32b36e87ae6927c85/django/core/handlers/base.py#L174
    """
    patch_function(m, "BaseHandler._get_response", _get_response_before)
    patch_function(m, "BaseHandler._get_response", _get_response_after)



"""
Helper function file, see funtion definition
"""

from aikido_zen.helpers.try_decode_as_jwt import try_decode_as_jwt
from aikido_zen.helpers.is_mapping import is_mapping
from aikido_zen.helpers.build_path_to_payload import build_path_to_payload
import aikido_zen.context as ctx


def extract_strings_from_user_input_cached(obj, source):
    """Use the cache to speed up getting user input"""
    context = ctx.get_current_context()

    if not context:
        #  context may not exist in all situations, in that cases the cache should be skipped
        return extract_strings_from_user_input(obj)

    if context.parsed_userinput and context.parsed_userinput.get(source):
        return context.parsed_userinput.get(source)
    res = extract_strings_from_user_input(obj)

    context.parsed_userinput[source] = res
    context.set_as_current_context()
    return res


def extract_strings_from_user_input(obj, path_to_payload=None):
    """
    Extracts strings from an object (user input)
    """
    if path_to_payload is None:
        path_to_payload = []

    results = {}

    if is_mapping(obj):
        #  Stringifying the dict and adding it as user input is resource intensive
        #  And in most cases shouldn't be necessary.
        for key, value in obj.items():
            results[key] = build_path_to_payload(path_to_payload)
            for k, v in extract_strings_from_user_input(
                value, path_to_payload + [{"type": "object", "key": key}]
            ).items():
                results[k] = v

    if isinstance(obj, (set, list, tuple)):
        #  Add the stringified array as well to the results, there might
        #  be accidental concatenation if the client expects a string but gets the array
        #  E.g. HTTP Parameter pollution
        results[str(obj)] = build_path_to_payload(path_to_payload)
        for i, value in enumerate(obj):
            for k, v in extract_strings_from_user_input(
                value, path_to_payload + [{"type": "array", "index": i}]
            ).items():
                results[k] = v

    if isinstance(obj, str):
        results[obj] = build_path_to_payload(path_to_payload)
        jwt = try_decode_as_jwt(obj)
        if jwt[0]:
            for k, v in extract_strings_from_user_input(
                jwt[1], path_to_payload + [{"type": "jwt"}]
            ).items():
                if k == "iss" or v.endswith("<jwt>.iss"):
                    # Do not add the issuer of the JWT as a string because it can contain a
                    # domain / url and produce false positives
                    continue
                results[k] = v

    return results



import pytest
import time
from .lru_cache import LRUCache


def test_lru_cache_creation():
    map = LRUCache(5, 1000)
    assert map.size == 0, "Size should be 0 initially"


def test_lru_cache_invalid_constructor_args():
    with pytest.raises(ValueError, match="Invalid max value"):
        LRUCache(-1, 1000)
    with pytest.raises(ValueError, match="Invalid ttl value"):
        LRUCache(100, -1)


def test_lru_cache_set_and_get_methods():
    map = LRUCache(5, 1000)

    map.set("key1", "value1")
    assert map.get("key1") == "value1", "Value should be retrieved correctly"

    map.set("key2", "value2")
    assert map.get("key2") == "value2", "Value should be retrieved correctly"

    assert map.size == 2, "Size should be 2 after adding two items"


def test_lru_cache_eviction_policy():
    map = LRUCache(2, 1000)

    map.set(1, "value1")
    map.set(2, "value2")
    assert map.size == 2, "Size should be 2 after adding two items"

    map.set(3, "value3")
    assert map.size == 2, "Size should be 2 after adding third item"
    assert map.get(1) is None, "First item should be evicted"
    assert map.get(2) == "value2", "Second item should still be present"
    assert map.get(3) == "value3", "Third item should be present"


def test_lru_cache_ttl_expiration():
    map = LRUCache(5, 100)

    map.set("key1", "value1")
    assert map.get("key1") == "value1", "Value should be retrieved correctly"

    # Wait for TTL to expire
    time.sleep(0.15)

    assert map.get("key1") is None, "Value should be None after TTL expiration"
    assert map.size == 0, "Size should be 0 after TTL expiration"


def test_lru_cache_clear_method():
    map = LRUCache(5, 1000)

    map.set("key1", "value1")
    map.set("key2", "value2")
    assert map.size == 2, "Size should be 2 after adding two items"

    map.clear()
    assert map.size == 0, "Size should be 0 after clearing"
    assert map.get("key1") is None, "Value should be None after clearing"
    assert map.get("key2") is None, "Value should be None after clearing"


def test_lru_cache_delete_method():
    map = LRUCache(5, 1000)

    map.set("key1", "value1")
    map.set("key2", "value2")
    assert map.size == 2, "Size should be 2 after adding two items"

    map.delete("key1")
    assert map.size == 1, "Size should be 1 after deleting one item"
    assert map.get("key1") is None, "Value should be None after deletion"
    assert (
        map.get("key2") == "value2"
    ), "Value should be retrieved correctly for remaining item"


def test_lru_cache_keys_method():
    map = LRUCache(5, 1000)

    map.set("key1", "value1")
    map.set("key2", "value2")
    map.set("key3", "value3")

    keys = map.keys()
    assert keys == ["key1", "key2", "key3"], "Keys should be retrieved correctly"



import pytest
from .get_ratelimited_endpoint import get_ratelimited_endpoint


@pytest.fixture
def endpoints():
    return [
        {
            "method": "POST",
            "route": "/api/login",
            "forceProtectionOff": False,
            "allowedIPAddresses": [],
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 3,
                "windowSizeInMS": 1000,
            },
        },
        {
            "method": "POST",
            "route": "/api/*",
            "forceProtectionOff": False,
            "allowedIPAddresses": [],
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 1,
                "windowSizeInMS": 1000,
            },
        },
        {
            "method": "GET",
            "route": "/",
            "forceProtectionOff": False,
            "allowedIPAddresses": [],
            "rateLimiting": {
                "enabled": False,
                "maxRequests": 3,
                "windowSizeInMS": 1000,
            },
        },
    ]


def test_returns_none_if_no_endpoints():
    assert get_ratelimited_endpoint([], "/api/login") is None


def test_returns_none_if_no_matching_endpoints(endpoints):
    assert get_ratelimited_endpoint([], "/nonexistent") is None


def test_returns_none_if_matching_but_not_enabled(endpoints):
    endpoints[0]["rateLimiting"]["enabled"] = False
    assert get_ratelimited_endpoint([endpoints[0]], "/api/login") is None


def test_returns_endpoint_if_matching_and_enabled(endpoints):
    result = get_ratelimited_endpoint(endpoints, "/api/login")
    assert result == endpoints[0]


def test_returns_endpoint_with_lowest_max_requests(endpoints):
    result = get_ratelimited_endpoint(endpoints, "/api/log*")
    assert result == endpoints[1]  # The one with maxRequests = 1


def test_returns_endpoint_with_smallest_window_size(endpoints):
    endpoints.append(
        {
            "method": "POST",
            "route": "/api/log*",
            "forceProtectionOff": False,
            "allowedIPAddresses": [],
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 3,
                "windowSizeInMS": 5000,
            },
        }
    )
    result = get_ratelimited_endpoint(endpoints, "/api/log*")
    assert (
        result["rateLimiting"]["windowSizeInMS"] == 5000
    )  # The one with the larger window size


def test_always_returns_exact_matches_first(endpoints):
    result = get_ratelimited_endpoint(endpoints, "/api/login")
    assert result == endpoints[0]  # Exact match should be returned first



import pytest
from aikido_zen.helpers.is_useful_route import is_useful_route


def test_not_found_or_method_not_allowed():
    assert is_useful_route(status_code=404, route="/", method="GET") is False
    assert is_useful_route(status_code=405, route="/", method="GET") is False


def test_discover_route_for_all_other_status_codes():
    assert is_useful_route(status_code=200, route="/", method="GET") is True
    assert is_useful_route(status_code=500, route="/", method="GET") is False
    assert is_useful_route(status_code=400, route="/", method="GET") is False
    assert is_useful_route(status_code=199, route="/", method="GET") is False
    assert is_useful_route(status_code=399, route="/", method="GET") is True
    assert is_useful_route(status_code=300, route="/", method="GET") is True
    assert is_useful_route(status_code=201, route="/", method="GET") is True


def test_not_discover_route_for_excluded_methods():
    assert is_useful_route(status_code=200, route="/", method="OPTIONS") is False
    assert is_useful_route(status_code=200, route="/", method="HEAD") is False


def test_not_discover_route_for_excluded_methods_with_other_status_codes():
    assert is_useful_route(status_code=404, route="/", method="OPTIONS") is False
    assert is_useful_route(status_code=405, route="/", method="HEAD") is False


def test_not_discover_static_files():
    assert (
        is_useful_route(status_code=200, route="/service-worker.js", method="GET")
        is False
    )
    assert (
        is_useful_route(
            status_code=200,
            route="/precache-manifest.10faec0bee24db502c8498078126dd53.js",
            method="POST",
        )
        is False
    )
    assert (
        is_useful_route(
            status_code=200, route="/img/icons/favicon-16x16.png", method="GET"
        )
        is False
    )
    assert (
        is_useful_route(status_code=200, route="/fonts/icomoon.ttf", method="GET")
        is False
    )


def test_allow_html_files():
    assert is_useful_route(status_code=200, route="/index.html", method="GET") is False
    assert (
        is_useful_route(status_code=200, route="/contact.html", method="GET") is False
    )


def test_allow_files_with_one_character_extension():
    assert is_useful_route(status_code=200, route="/a.a", method="GET") is True


def test_allow_files_with_six_or_more_character_extensions():
    assert is_useful_route(status_code=200, route="/a.aaaaaa", method="GET") is True
    assert is_useful_route(status_code=200, route="/a.aaaaaaa", method="GET") is True


def test_ignore_files_that_end_with_properties():
    assert (
        is_useful_route(status_code=200, route="/file.properties", method="GET")
        is False
    )
    assert (
        is_useful_route(
            status_code=200, route="/directory/file.properties", method="GET"
        )
        is False
    )


def test_ignore_files_or_directories_that_start_with_dot():
    assert is_useful_route(status_code=200, route="/.env", method="GET") is False
    assert (
        is_useful_route(status_code=200, route="/.aws/credentials", method="GET")
        is False
    )
    assert (
        is_useful_route(status_code=200, route="/directory/.gitconfig", method="GET")
        is False
    )
    assert (
        is_useful_route(status_code=200, route="/hello/.gitignore/file", method="GET")
        is False
    )


def test_ignore_files_that_end_with_php():
    assert is_useful_route(status_code=200, route="/file.php", method="GET") is False
    assert (
        is_useful_route(
            status_code=200, route="/app_dev.php/_profiler/phpinfo", method="GET"
        )
        is False
    )


def test_allow_well_known_directory():
    assert is_useful_route(status_code=200, route="/.well-known", method="GET") is False
    assert (
        is_useful_route(
            status_code=200, route="/.well-known/change-password", method="GET"
        )
        is True
    )
    assert (
        is_useful_route(
            status_code=200, route="/.well-known/security.txt", method="GET"
        )
        is False
    )


def test_ignore_certain_strings():
    assert (
        is_useful_route(
            status_code=200, route="/cgi-bin/luci/;stok=/locale", method="GET"
        )
        is False
    )
    assert (
        is_useful_route(status_code=200, route="/whatever/cgi-bin", method="GET")
        is False
    )


def test_ignore_fonts():
    assert (
        is_useful_route(status_code=200, route="/fonts/icomoon.ttf", method="GET")
        is False
    )
    assert (
        is_useful_route(status_code=200, route="/fonts/icomoon.woff", method="GET")
        is False
    )
    assert (
        is_useful_route(status_code=200, route="/fonts/icomoon.woff2", method="GET")
        is False
    )


def test_ignore_files_that_end_with_config():
    assert (
        is_useful_route(
            status_code=200,
            route="/blog/App_Config/ConnectionStrings.config",
            method="GET",
        )
        is False
    )


def test_allow_redirects():
    assert is_useful_route(status_code=301, route="/", method="GET") is True
    assert is_useful_route(status_code=302, route="/", method="GET") is True
    assert is_useful_route(status_code=303, route="/", method="GET") is True
    assert is_useful_route(status_code=307, route="/", method="GET") is True
    assert is_useful_route(status_code=308, route="/", method="GET") is True



"""
Sink module for `http`
"""

from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import before, after, patch_function, on_import
from aikido_zen.vulnerabilities.ssrf.handle_http_response import (
    handle_http_response,
)
from aikido_zen.helpers.try_parse_url import try_parse_url


@before
def _putrequest(func, instance, args, kwargs):
    # putrequest(...) is called with path argument, store this on the HTTPConnection
    # instance for later use in the getresponse(...) function.
    path = get_argument(args, kwargs, 1, "path")
    setattr(instance, "_aikido_var_path", path)


@after
def _getresponse(func, instance, args, kwargs, return_value):
    path = getattr(instance, "_aikido_var_path")
    source_url = try_parse_url(f"http://{instance.host}:{instance.port}{path}")

    register_call("HTTPConnection.getresponse", "outgoing_http_op")

    handle_http_response(http_response=return_value, source=source_url)


@on_import("http.client")
def patch(m):
    """
    patching module http.client
    - patches HTTPConnection.putrequest -> stores path
    - patches HTTPConnection.getresponse -> handles response object
    """
    patch_function(m, "HTTPConnection.putrequest", _putrequest)
    patch_function(m, "HTTPConnection.getresponse", _getresponse)



"""
Starlette wrapping is subdivided in two parts :
- starlette.applications : Wraps __call__ on Starlette class to run "init" stage.
- starlette.routing : request_response function : Run pre_response code and 
    also runs post_response code after getting response from user function.

Folder also includes helper functions : 
- extract_data_from_request, which will extract the data from a request object safely,
    e.g. body, json, form. This also saves it inside the current context.
"""

import aikido_zen.sources.starlette.starlette_applications
import aikido_zen.sources.starlette.starlette_routing



"""
imds.py file, exports :
is_imds_ip_address, is_trusted_hostname, resolves_to_imds_ip
"""


class BlockList:
    """A list of IP's that shouldn't be accessed"""

    def __init__(self):
        self.blocked_addresses = {"ipv4": set(), "ipv6": set()}

    def add_address(self, address, address_type):
        """Add an address to this list"""
        if address_type in self.blocked_addresses:
            self.blocked_addresses[address_type].add(address)

    def check(self, address, address_type=None):
        """Check if the IP is on the list"""
        if address_type:
            return address in self.blocked_addresses.get(address_type, set())
        return any(
            address in addresses for addresses in self.blocked_addresses.values()
        )


# Create an instance of BlockList
imds_addresses = BlockList()

# Block the IP addresses used by AWS EC2 instances for IMDS
imds_addresses.add_address("169.254.169.254", "ipv4")
imds_addresses.add_address("fd00:ec2::254", "ipv6")


def is_imds_ip_address(ip):
    """Checks if the IP is an imds ip"""
    return imds_addresses.check(ip) or imds_addresses.check(ip, "ipv6")


# Trusted hostnames for Google Cloud
trusted_hosts = ["metadata.google.internal", "metadata.goog"]


def is_trusted_hostname(hostname):
    """Checks if this hostname is trusted"""
    return hostname in trusted_hosts


def resolves_to_imds_ip(resolved_ip_addresses, hostname):
    """
    returns a boolean, true if the IP is an imds ip
    """
    #  Allow access to Google Cloud metadata service as you need to set specific headers to access it
    #  We don't want to block legitimate requests
    if is_trusted_hostname(hostname):
        return False
    return any(is_imds_ip_address(ip) for ip in resolved_ip_addresses)



import pytest
from unittest.mock import patch
import aikido_zen.sinks.mysqlclient
from aikido_zen.background_process.comms import reset_comms

kind = "sql_injection"
op = "MySQLdb.connections.query"


@pytest.fixture
def database_conn():
    import MySQLdb

    return MySQLdb.connect(host="127.0.0.1", user="user", passwd="password", db="db")


def test_cursor_execute(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        query = "SELECT * FROM dogs"
        cursor.execute(query)
        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "mysql"

        mock_run_vulnerability_scan.assert_called_once()
        cursor.fetchall()
        mock_run_vulnerability_scan.assert_called_once()

        cursor.close()
        database_conn.close()
        mock_run_vulnerability_scan.assert_called_once()


def test_cursor_execute_no_args(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        dogname = "Doggo"
        isadmin = "TRUE"
        query = f'INSERT INTO dogs (dog_name, isAdmin) VALUES ("{dogname}", {isadmin})'
        cursor.execute(query)
        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert (
            called_with_args[0]
            == 'INSERT INTO dogs (dog_name, isAdmin) VALUES ("Doggo", TRUE)'
        )
        assert called_with_args[1] == "mysql"

        mock_run_vulnerability_scan.assert_called_once()
        cursor.fetchall()
        mock_run_vulnerability_scan.assert_called_once()

        cursor.close()
        database_conn.close()
        mock_run_vulnerability_scan.assert_called_once()


def test_cursor_executemany(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        data = [
            ("Doggy", False),
            ("Doggy 2", True),
            ("Dogski", True),
        ]
        cursor.executemany("INSERT INTO dogs (dog_name, isAdmin) VALUES (%s, %s)", data)
        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert (
            called_with_args[0]
            == "INSERT INTO dogs (dog_name, isAdmin) VALUES (%s, %s)"
        )
        assert called_with_args[1] == "mysql"
        mock_run_vulnerability_scan.assert_called_once()
        database_conn.commit()
        mock_run_vulnerability_scan.assert_called_once()

        cursor.close()
        database_conn.close()
        mock_run_vulnerability_scan.assert_called_once()


def test_cursor_execute_with_fstring(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        data = ("doggy",)
        table_name = "dogs"
        value_2 = "1"
        cursor.execute(
            f"INSERT INTO {table_name} (dog_name, isAdmin) VALUES (%s, {value_2})", data
        )
        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert (
            called_with_args[0] == "INSERT INTO dogs (dog_name, isAdmin) VALUES (%s, 1)"
        )
        assert called_with_args[1] == "mysql"
        mock_run_vulnerability_scan.assert_called_once()
        database_conn.commit()
        mock_run_vulnerability_scan.assert_called_once()

        cursor.close()
        database_conn.close()
        mock_run_vulnerability_scan.assert_called_once()
